"""Dependency provider traversal tests."""

from dependency_injector import providers


def test_traversal():
    provider = providers.Dependency()
    all_providers = list(provider.traverse())
    assert len(all_providers) == 0


def test_traversal_default():
    another_provider = providers.Provider()
    provider = providers.Dependency(default=another_provider)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 1
    assert another_provider in all_providers


def test_traversal_overriding():
    provider1 = providers.Provider()

    provider2 = providers.Provider()
    provider2.override(provider1)

    provider = providers.Dependency()
    provider.override(provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers



"""FactoryAggregate provider async mode tests."""

from dependency_injector import providers
from pytest import mark


@mark.asyncio
async def test_async_mode():
    object1 = object()
    object2 = object()

    async def _get_object1():
        return object1

    def _get_object2():
        return object2

    provider = providers.FactoryAggregate(
        object1=providers.Factory(_get_object1),
        object2=providers.Factory(_get_object2),
    )

    assert provider.is_async_mode_undefined() is True

    created_object1 = await provider("object1")
    assert created_object1 is object1
    assert provider.is_async_mode_enabled() is True

    created_object2 = await provider("object2")
    assert created_object2 is object2



"""ProvidedInstance provider traversal tests."""

from dependency_injector import providers


def test_traverse():
    provider1 = providers.Provider()
    provider = provider1.provided

    all_providers = list(provider.traverse())

    assert len(all_providers) == 1
    assert provider1 in all_providers


def test_traverse_overridden():
    provider1 = providers.Provider()
    provider2 = providers.Provider()

    provider = provider1.provided
    provider.override(provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers



"""DelegatedCallable provider tests."""

from dependency_injector import providers

from .common import example


def test_inheritance():
    assert isinstance(providers.DelegatedCallable(example), providers.Callable)


def test_is_provider():
    assert providers.is_provider(providers.DelegatedCallable(example)) is True


def test_is_delegated_provider():
    provider = providers.DelegatedCallable(example)
    assert providers.is_delegated(provider) is True


def test_repr():
    provider = providers.DelegatedCallable(example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "DelegatedCallable({0}) at {1}>".format(repr(example), hex(id(provider)))
    )



"""Aiohttp extension module."""

from __future__ import absolute_import

import functools
import warnings

from dependency_injector import providers


warnings.warn(
    'Module "dependency_injector.ext.aiohttp" is deprecated since '
    'version 4.0.0. Use "dependency_injector.wiring" module instead.',
    category=DeprecationWarning,
)


class Application(providers.Singleton):
    """Aiohttp application provider."""


class Extension(providers.Singleton):
    """Aiohttp extension provider."""


class Middleware(providers.DelegatedCallable):
    """Aiohttp middleware provider."""

    __middleware_version__ = 1


class MiddlewareFactory(providers.Factory):
    """Aiohttp middleware factory provider."""


class View(providers.Callable):
    """Aiohttp view provider."""

    def as_view(self):
        """Return aiohttp view function."""
        @functools.wraps(self.provides)
        async def _view(request, *args, **kwargs):
            return await self.__call__(request, *args, **kwargs)
        return _view


class ClassBasedView(providers.Factory):
    """Aiohttp class-based view provider."""

    def as_view(self):
        """Return aiohttp view function."""
        async def _view(request, *args, **kwargs):
            return await self.__call__(request, *args, **kwargs)
        return _view



"""Resource provider traversal tests."""

from dependency_injector import providers


def test_traverse():
    provider = providers.Resource(dict)
    all_providers = list(provider.traverse())
    assert len(all_providers) == 0


def test_traverse_args():
    provider1 = providers.Object("bar")
    provider2 = providers.Object("baz")
    provider = providers.Resource(list, "foo", provider1, provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers


def test_traverse_kwargs():
    provider1 = providers.Object("bar")
    provider2 = providers.Object("baz")
    provider = providers.Resource(dict, foo="foo", bar=provider1, baz=provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers


def test_traverse_overridden():
    provider1 = providers.Resource(list)
    provider2 = providers.Resource(tuple)

    provider = providers.Resource(dict, "foo")
    provider.override(provider1)
    provider.override(provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers


def test_traverse_provides():
    provider1 = providers.Callable(list)

    provider = providers.Resource(provider1)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 1
    assert provider1 in all_providers




from typing import Tuple, Any, List

from dependency_injector import providers


# Test 1: to check the return type (class)
provider1 = providers.List(
    providers.Factory(object),
    providers.Factory(object),
)
var1: List[Any] = provider1()


# Test 2: to check the .args attributes
provider2 = providers.List(
    providers.Factory(object),
    providers.Factory(object),
)
args2: Tuple[Any] = provider2.args

# Test 3: to check the provided instance interface
provider3 = providers.List(
    providers.Factory(object),
    providers.Factory(object),
)
provided3: providers.ProvidedInstance = provider3.provided
attr_getter3: providers.AttributeGetter = provider3.provided.attr
item_getter3: providers.ItemGetter = provider3.provided["item"]
method_caller3: providers.MethodCaller = provider3.provided.method.call(123, arg=324)

# Test 4: to check the return type with await
provider4 = providers.List(
    providers.Factory(object),
    providers.Factory(object),
)
async def _async4() -> None:
    var1: List[Any] = await provider4()  # type: ignore
    var2: List[Any] = await provider4.async_()



from flask import Flask, jsonify, request, current_app, session, g
# from flask import _request_ctx_stack, _app_ctx_stack
from dependency_injector import containers, providers
from dependency_injector.wiring import inject, Provide

# This is here for testing wiring bypasses these objects without crashing
request, current_app, session, g  # noqa
# _request_ctx_stack, _app_ctx_stack  # noqa


class Service:
    def process(self) -> str:
        return "OK"


class Container(containers.DeclarativeContainer):

    service = providers.Factory(Service)


app = Flask(__name__)


@app.route("/")
@inject
def index(service: Service = Provide[Container.service]):
    result = service.process()
    return jsonify({"result": result})


container = Container()
container.wire(modules=[__name__])



"""Main wiring tests."""

from decimal import Decimal

from dependency_injector import errors
from dependency_injector.wiring import Closing, Provide, Provider, wire
from pytest import fixture, mark, raises

from samples.wiringstringids import module, package, resourceclosing
from samples.wiringstringids.service import Service
from samples.wiringstringids.container import Container, SubContainer


@fixture(autouse=True)
def container():
    container = Container(config={"a": {"b": {"c": 10}}})
    container.wire(
        modules=[module],
        packages=[package],
    )
    yield container
    container.unwire()


@fixture
def subcontainer():
    container = SubContainer()
    container.wire(
        modules=[module],
        packages=[package],
    )
    yield container
    container.unwire()


@fixture
def resourceclosing_container():
    container = resourceclosing.Container()
    container.wire(modules=[resourceclosing])
    yield container
    container.unwire()


def test_package_lookup():
    from samples.wiringstringids.package import test_package_function
    service = test_package_function()
    assert isinstance(service, Service)


def test_package_subpackage_lookup():
    from samples.wiringstringids.package.subpackage import test_package_function
    service = test_package_function()
    assert isinstance(service, Service)


def test_package_submodule_lookup():
    from samples.wiringstringids.package.subpackage.submodule import test_function
    service = test_function()
    assert isinstance(service, Service)


def test_module_attributes_wiring():
    assert isinstance(module.service, Service)
    assert isinstance(module.service_provider(), Service)
    assert isinstance(module.undefined, Provide)


def test_module_attribute_wiring_with_invalid_marker(container: Container):
    from samples.wiringstringids import module_invalid_attr_injection
    with raises(Exception, match="Unknown type of marker {0}".format(module_invalid_attr_injection.service)):
        container.wire(modules=[module_invalid_attr_injection])


def test_class_wiring():
    test_class_object = module.TestClass()
    assert isinstance(test_class_object.service, Service)


def test_class_wiring_context_arg(container: Container):
    test_service = container.service()
    test_class_object = module.TestClass(service=test_service)
    assert test_class_object.service is test_service


def test_class_method_wiring():
    test_class_object = module.TestClass()
    service = test_class_object.method()
    assert isinstance(service, Service)


def test_class_classmethod_wiring():
    service = module.TestClass.class_method()
    assert isinstance(service, Service)


def test_instance_classmethod_wiring():
    instance = module.TestClass()
    service = instance.class_method()
    assert isinstance(service, Service)


def test_class_staticmethod_wiring():
    service = module.TestClass.static_method()
    assert isinstance(service, Service)


def test_instance_staticmethod_wiring():
    instance = module.TestClass()
    service = instance.static_method()
    assert isinstance(service, Service)


def test_class_attribute_wiring():
    assert isinstance(module.TestClass.service, Service)
    assert isinstance(module.TestClass.service_provider(), Service)
    assert isinstance(module.TestClass.undefined, Provide)


def test_function_wiring():
    service = module.test_function()
    assert isinstance(service, Service)


def test_function_wiring_context_arg(container: Container):
    test_service = container.service()
    service = module.test_function(service=test_service)
    assert service is test_service


def test_function_wiring_provider():
    service = module.test_function_provider()
    assert isinstance(service, Service)


def test_function_wiring_provider_context_arg(container: Container):
    test_service = container.service()
    service = module.test_function_provider(service_provider=lambda: test_service)
    assert service is test_service


def test_configuration_option():
    (
        value_int,
        value_float,
        value_str,
        value_decimal,
        value_required,
        value_required_int,
        value_required_float,
        value_required_str,
        value_required_decimal,
    ) = module.test_config_value()

    assert value_int == 10
    assert value_float == 10.0
    assert value_str == "10"
    assert value_decimal == Decimal(10)
    assert value_required == 10
    assert value_required_int == 10
    assert value_required_float == 10.0
    assert value_required_str == "10"
    assert value_required_decimal == Decimal(10)


def test_configuration_option_required_undefined(container: Container):
    container.config.reset_override()
    with raises(errors.Error, match="Undefined configuration option \"config.a.b.c\""):
        module.test_config_value_required_undefined()


def test_provide_provider():
    service = module.test_provide_provider()
    assert isinstance(service, Service)


def test_provider_provider():
    service = module.test_provider_provider()
    assert isinstance(service, Service)


def test_provided_instance(container: Container):
    class TestService:
        foo = {"bar": lambda: 10}

    with container.service.override(TestService()):
        some_value = module.test_provided_instance()
    assert some_value == 10


def test_subcontainer():
    some_value = module.test_subcontainer_provider()
    assert some_value == 1


def test_config_invariant(container: Container):
    config = {
        "option": {
            "a": 1,
            "b": 2,
        },
        "switch": "a",
    }
    container.config.from_dict(config)

    value_default = module.test_config_invariant()
    assert value_default == 1

    with container.config.switch.override("a"):
        value_a = module.test_config_invariant()
    assert value_a == 1

    with container.config.switch.override("b"):
        value_b = module.test_config_invariant()
    assert value_b == 2


def test_wire_with_class_error():
    with raises(Exception):
        wire(
            container=Container,
            modules=[module],
        )


def test_unwire_function(container: Container):
    container.unwire()
    assert isinstance(module.test_function(), Provide)


def test_unwire_class(container: Container):
    container.unwire()
    test_class_object = module.TestClass()
    assert isinstance(test_class_object.service, Provide)


def test_unwire_class_method(container: Container):
    container.unwire()
    test_class_object = module.TestClass()
    assert isinstance(test_class_object.method(), Provide)


def test_unwire_package_function(container: Container):
    container.unwire()
    from samples.wiringstringids.package.subpackage.submodule import test_function
    assert isinstance(test_function(), Provide)


def test_unwire_package_function_by_reference(container: Container):
    from samples.wiringstringids.package.subpackage import submodule
    container.unwire()
    assert isinstance(submodule.test_function(), Provide)


def test_unwire_module_attributes(container: Container):
    container.unwire()
    assert isinstance(module.service, Provide)
    assert isinstance(module.service_provider, Provider)
    assert isinstance(module.undefined, Provide)


def test_unwire_class_attributes(container: Container):
    container.unwire()
    assert isinstance(module.TestClass.service, Provide)
    assert isinstance(module.TestClass.service_provider, Provider)
    assert isinstance(module.TestClass.undefined, Provide)


@mark.usefixtures("subcontainer")
def test_wire_multiple_containers():
    service, some_value = module.test_provide_from_different_containers()
    assert isinstance(service, Service)
    assert some_value == 1


@mark.usefixtures("resourceclosing_container")
def test_closing_resource():
    resourceclosing.Service.reset_counter()

    result_1 = resourceclosing.test_function()
    assert isinstance(result_1, resourceclosing.Service)
    assert result_1.init_counter == 1
    assert result_1.shutdown_counter == 1

    result_2 = resourceclosing.test_function()
    assert isinstance(result_2, resourceclosing.Service)
    assert result_2.init_counter == 2
    assert result_2.shutdown_counter == 2

    assert result_1 is not result_2


@mark.usefixtures("resourceclosing_container")
def test_closing_dependency_resource():
    resourceclosing.Service.reset_counter()

    result_1 = resourceclosing.test_function_dependency()
    assert isinstance(result_1, resourceclosing.FactoryService)
    assert result_1.service.init_counter == 1
    assert result_1.service.shutdown_counter == 1

    result_2 = resourceclosing.test_function_dependency()
    assert isinstance(result_2, resourceclosing.FactoryService)
    assert result_2.service.init_counter == 2
    assert result_2.service.shutdown_counter == 2

    assert result_1 is not result_2


@mark.usefixtures("resourceclosing_container")
def test_closing_resource_bypass_marker_injection():
    resourceclosing.Service.reset_counter()

    result_1 = resourceclosing.test_function(service=Closing[Provide["service"]])
    assert isinstance(result_1, resourceclosing.Service)
    assert result_1.init_counter == 1
    assert result_1.shutdown_counter == 1

    result_2 = resourceclosing.test_function(service=Closing[Provide["service"]])
    assert isinstance(result_2, resourceclosing.Service)
    assert result_2.init_counter == 2
    assert result_2.shutdown_counter == 2

    assert result_1 is not result_2


@mark.usefixtures("resourceclosing_container")
def test_closing_resource_context():
    resourceclosing.Service.reset_counter()
    service = resourceclosing.Service()

    result_1 = resourceclosing.test_function(service=service)
    assert result_1 is service
    assert result_1.init_counter == 0
    assert result_1.shutdown_counter == 0

    result_2 = resourceclosing.test_function(service=service)
    assert result_2 is service
    assert result_2.init_counter == 0
    assert result_2.shutdown_counter == 0


def test_class_decorator():
    service = module.test_class_decorator()
    assert isinstance(service, Service)


def test_container():
    service = module.test_container()
    assert isinstance(service, Service)


def test_bypass_marker_injection():
    service = module.test_function(service=Provide["service"])
    assert isinstance(service, Service)



from typing import Type

from dependency_injector import providers


class Animal:
    ...


class Cat(Animal):

    def __init__(self, *_, **__): ...


# Test 1: to check the return type
provider1 = providers.Dependency(instance_of=Animal)
provider1.override(providers.Factory(Cat))
var1: Animal = provider1()

# Test 2: to check the return type
provider2 = providers.Dependency(instance_of=Animal)
var2: Type[Animal] = provider2.instance_of

# Test 3: to check the return type with await
provider3 = providers.Dependency(instance_of=Animal)
async def _async3() -> None:
    var1: Animal = await provider3()  # type: ignore
    var2: Animal = await provider3.async_()



import sys

from fastapi import FastAPI, Depends
from fastapi import Request  # See: https://github.com/ets-labs/python-dependency-injector/issues/398
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from dependency_injector import containers, providers
from dependency_injector.wiring import inject, Provide


class Service:
    async def process(self) -> str:
        return "Ok"


class Container(containers.DeclarativeContainer):

    service = providers.Factory(Service)


app = FastAPI()
security = HTTPBasic()


@app.api_route("/")
@inject
async def index(service: Service = Depends(Provide[Container.service])):
    result = await service.process()
    return {"result": result}


@app.get("/auth")
@inject
def read_current_user(
        credentials: HTTPBasicCredentials = Depends(security)
):
    return {"username": credentials.username, "password": credentials.password}


container = Container()
container.wire(modules=[sys.modules[__name__]])



"""Tests for specifying wiring config in the container."""

from dependency_injector import containers
from dependency_injector.wiring import Provide
from pytest import fixture, mark

from samples.wiring import module
from samples.wiring.service import Service
from samples.wiring.container import Container


@fixture(autouse=True)
def container(wiring_config: containers.WiringConfiguration):
    original_wiring_config = Container.wiring_config
    Container.wiring_config = wiring_config
    container = Container()
    yield container
    container.unwire()
    Container.wiring_config = original_wiring_config


@mark.parametrize(
    "wiring_config",
    [
        containers.WiringConfiguration(
            modules=["samples.wiring.module"],
            packages=["samples.wiring.package"],
        ),
    ],
)
def test_absolute_names():
    service = module.test_function()
    assert isinstance(service, Service)

    from samples.wiring.package.subpackage.submodule import test_function
    service = test_function()
    assert isinstance(service, Service)


@mark.parametrize(
    "wiring_config",
    [
        containers.WiringConfiguration(
            modules=[".module"],
            packages=[".package"],
            from_package="samples.wiring",
        ),
    ],
)
def test_relative_names_with_explicit_package():
    service = module.test_function()
    assert isinstance(service, Service)

    from samples.wiring.package.subpackage.submodule import test_function
    service = test_function()
    assert isinstance(service, Service)


@mark.parametrize(
    "wiring_config",
    [
        containers.WiringConfiguration(
            modules=[".module"],
            packages=[".package"],
        ),
    ],
)
def test_relative_names_with_auto_package():
    service = module.test_function()
    assert isinstance(service, Service)

    from samples.wiring.package.subpackage.submodule import test_function
    service = test_function()
    assert isinstance(service, Service)


@mark.parametrize(
    "wiring_config",
    [
        containers.WiringConfiguration(
            modules=[".module"],
            auto_wire=False,
        ),
    ],
)
def test_auto_wire_disabled(container: Container):
    service = module.test_function()
    assert isinstance(service, Provide)

    container.wire()
    service = module.test_function()
    assert isinstance(service, Service)



"""Selector provider traversal tests."""

from dependency_injector import providers


def test_traverse():
    switch = lambda: "provider1"
    provider1 = providers.Callable(list)
    provider2 = providers.Callable(dict)

    provider = providers.Selector(
        switch,
        provider1=provider1,
        provider2=provider2,
    )

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers


def test_traverse_switch():
    switch = providers.Callable(lambda: "provider1")
    provider1 = providers.Callable(list)
    provider2 = providers.Callable(dict)

    provider = providers.Selector(
        switch,
        provider1=provider1,
        provider2=provider2,
    )

    all_providers = list(provider.traverse())

    assert len(all_providers) == 3
    assert switch in all_providers
    assert provider1 in all_providers
    assert provider2 in all_providers


def test_traverse_overridden():
    provider1 = providers.Callable(list)
    provider2 = providers.Callable(dict)
    selector1 = providers.Selector(lambda: "provider1", provider1=provider1)

    provider = providers.Selector(
        lambda: "provider2",
        provider2=provider2,
    )
    provider.override(selector1)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 3
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert selector1 in all_providers



"""Main container class tests."""

import collections

from dependency_injector import containers, providers, errors
from pytest import raises


class ContainerA(containers.DeclarativeContainer):
    p11 = providers.Provider()
    p12 = providers.Provider()


class ContainerB(ContainerA):
    p21 = providers.Provider()
    p22 = providers.Provider()


class ContainerC(ContainerB):
    p31 = providers.Provider()
    p32 = providers.Provider()


def test_providers_attribute():
    assert ContainerA.providers == dict(p11=ContainerA.p11, p12=ContainerA.p12)
    assert ContainerB.providers == dict(
        p11=ContainerA.p11,
        p12=ContainerA.p12,
        p21=ContainerB.p21,
        p22=ContainerB.p22,
    )
    assert ContainerC.providers == dict(
        p11=ContainerA.p11,
        p12=ContainerA.p12,
        p21=ContainerB.p21,
        p22=ContainerB.p22,
        p31=ContainerC.p31,
        p32=ContainerC.p32,
    )


def test_providers_attribute_with_redefinition():
    p1 = providers.Provider()
    p2 = providers.Provider()

    class ContainerA2(ContainerA):
        p11 = p1
        p12 = p2

    assert ContainerA.providers == {
        "p11": ContainerA.p11,
        "p12": ContainerA.p12,
    }
    assert ContainerA2.providers == {
        "p11": p1,
        "p12": p2,
    }


def test_cls_providers_attribute():
    assert ContainerA.cls_providers == dict(p11=ContainerA.p11, p12=ContainerA.p12)
    assert ContainerB.cls_providers == dict(p21=ContainerB.p21, p22=ContainerB.p22)
    assert ContainerC.cls_providers == dict(p31=ContainerC.p31, p32=ContainerC.p32)


def test_inherited_providers_attribute():
    assert ContainerA.inherited_providers == dict()
    assert ContainerB.inherited_providers == dict(p11=ContainerA.p11, p12=ContainerA.p12)
    assert ContainerC.inherited_providers == dict(
        p11=ContainerA.p11,
        p12=ContainerA.p12,
        p21=ContainerB.p21,
        p22=ContainerB.p22,
    )


def test_dependencies_attribute():
    class ContainerD(ContainerC):
        p41 = providers.Dependency()
        p42 = providers.DependenciesContainer()

    class ContainerE(ContainerD):
        p51 = providers.Dependency()
        p52 = providers.DependenciesContainer()

    assert ContainerD.dependencies == {
        "p41": ContainerD.p41,
        "p42": ContainerD.p42,
    }
    assert ContainerE.dependencies == {
        "p41": ContainerD.p41,
        "p42": ContainerD.p42,
        "p51": ContainerE.p51,
        "p52": ContainerE.p52,
    }


def test_set_get_del_providers():
    a_p13 = providers.Provider()
    b_p23 = providers.Provider()

    ContainerA.p13 = a_p13
    ContainerB.p23 = b_p23

    assert ContainerA.providers == dict(
        p11=ContainerA.p11,
        p12=ContainerA.p12,
        p13=a_p13,
    )
    assert ContainerB.providers == dict(
        p11=ContainerA.p11,
        p12=ContainerA.p12,
        p21=ContainerB.p21,
        p22=ContainerB.p22,
        p23=b_p23,
    )

    assert ContainerA.cls_providers == dict(
        p11=ContainerA.p11,
        p12=ContainerA.p12,
        p13=a_p13,
    )
    assert ContainerB.cls_providers == dict(
        p21=ContainerB.p21,
        p22=ContainerB.p22,
        p23=b_p23,
    )

    del ContainerA.p13
    del ContainerB.p23

    assert ContainerA.providers == dict(p11=ContainerA.p11, p12=ContainerA.p12)
    assert ContainerB.providers == dict(
        p11=ContainerA.p11,
        p12=ContainerA.p12,
        p21=ContainerB.p21,
        p22=ContainerB.p22,
    )

    assert ContainerA.cls_providers == dict(p11=ContainerA.p11, p12=ContainerA.p12)
    assert ContainerB.cls_providers == dict(p21=ContainerB.p21, p22=ContainerB.p22)


def test_declare_with_valid_provider_type():
    class _Container(containers.DeclarativeContainer):
        provider_type = providers.Object
        px = providers.Object(object())

    assert isinstance(_Container.px, providers.Object)


def test_declare_with_invalid_provider_type():
    with raises(errors.Error):
        class _Container(containers.DeclarativeContainer):
            provider_type = providers.Object
            px = providers.Provider()


def test_seth_valid_provider_type():
    class _Container(containers.DeclarativeContainer):
        provider_type = providers.Object

    _Container.px = providers.Object(object())

    assert isinstance(_Container.px, providers.Object)

def test_set_invalid_provider_type():
    class _Container(containers.DeclarativeContainer):
        provider_type = providers.Object

    with raises(errors.Error):
        _Container.px = providers.Provider()


def test_override():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer1(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer2(containers.DeclarativeContainer):
        p11 = providers.Provider()
        p12 = providers.Provider()

    _Container.override(_OverridingContainer1)
    _Container.override(_OverridingContainer2)

    assert _Container.overridden == (_OverridingContainer1, _OverridingContainer2)
    assert _Container.p11.overridden == (_OverridingContainer1.p11, _OverridingContainer2.p11)


def test_override_with_it():
    with raises(errors.Error):
        ContainerA.override(ContainerA)


def test_override_with_parent():
    with raises(errors.Error):
        ContainerB.override(ContainerA)


def test_override_decorator():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Provider()

    @containers.override(_Container)
    class _OverridingContainer1(containers.DeclarativeContainer):
        p11 = providers.Provider()

    @containers.override(_Container)
    class _OverridingContainer2(containers.DeclarativeContainer):
        p11 = providers.Provider()
        p12 = providers.Provider()

    assert _Container.overridden == (_OverridingContainer1, _OverridingContainer2)
    assert _Container.p11.overridden == (_OverridingContainer1.p11, _OverridingContainer2.p11)


def test_reset_last_overriding():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer1(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer2(containers.DeclarativeContainer):
        p11 = providers.Provider()
        p12 = providers.Provider()

    _Container.override(_OverridingContainer1)
    _Container.override(_OverridingContainer2)
    _Container.reset_last_overriding()

    assert _Container.overridden == (_OverridingContainer1,)
    assert _Container.p11.overridden == (_OverridingContainer1.p11,)


def test_reset_last_overriding_when_not_overridden():
    with raises(errors.Error):
        ContainerA.reset_last_overriding()


def test_reset_override():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer1(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer2(containers.DeclarativeContainer):
        p11 = providers.Provider()
        p12 = providers.Provider()

    _Container.override(_OverridingContainer1)
    _Container.override(_OverridingContainer2)
    _Container.reset_override()

    assert _Container.overridden == tuple()
    assert _Container.p11.overridden == tuple()


def test_copy():
    @containers.copy(ContainerA)
    class _Container1(ContainerA):
        pass

    @containers.copy(ContainerA)
    class _Container2(ContainerA):
        pass

    assert ContainerA.p11 is not _Container1.p11
    assert ContainerA.p12 is not _Container1.p12

    assert ContainerA.p11 is not _Container2.p11
    assert ContainerA.p12 is not _Container2.p12

    assert _Container1.p11 is not _Container2.p11
    assert _Container1.p12 is not _Container2.p12


def test_copy_with_replacing():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Object(0)
        p12 = providers.Factory(dict, p11=p11)

    @containers.copy(_Container)
    class _Container1(_Container):
        p11 = providers.Object(1)
        p13 = providers.Object(11)

    @containers.copy(_Container)
    class _Container2(_Container):
        p11 = providers.Object(2)
        p13 = providers.Object(22)

    assert _Container.p11 is not _Container1.p11
    assert _Container.p12 is not _Container1.p12

    assert _Container.p11 is not _Container2.p11
    assert _Container.p12 is not _Container2.p12

    assert _Container1.p11 is not _Container2.p11
    assert _Container1.p12 is not _Container2.p12

    assert _Container.p12() == {"p11": 0}
    assert _Container1.p12() == {"p11": 1}
    assert _Container2.p12() == {"p11": 2}

    assert _Container1.p13() == 11
    assert _Container2.p13() == 22


def test_copy_with_parent_dependency():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/477
    class Base(containers.DeclarativeContainer):
        p11 = providers.Object(0)
        p12 = providers.Factory(dict, p11=p11)

    @containers.copy(Base)
    class New(Base):
        p13 = providers.Factory(dict, p12=Base.p12)

    new1 = New()
    new2 = New(p11=1)
    new3 = New(p11=2)

    assert new1.p13() == {"p12": {"p11": 0}}
    assert new2.p13() == {"p12": {"p11": 1}}
    assert new3.p13() == {"p12": {"p11": 2}}


def test_copy_with_replacing_subcontainer_providers():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/374
    class X(containers.DeclarativeContainer):
        foo = providers.Dependency(instance_of=str)

    def build_x():
        return X(foo="1")

    class A(containers.DeclarativeContainer):
        x = providers.DependenciesContainer(**X.providers)
        y = x.foo

    @containers.copy(A)
    class B1(A):
        x = providers.Container(build_x)

    b1 = B1()

    assert b1.y() == "1"


def test_containers_attribute():
    class Container(containers.DeclarativeContainer):
        class Container1(containers.DeclarativeContainer):
            pass

        class Container2(containers.DeclarativeContainer):
            pass

        Container3 = containers.DynamicContainer()

    assert Container.containers == dict(
        Container1=Container.Container1,
        Container2=Container.Container2,
        Container3=Container.Container3,
    )


def test_init_with_overriding_providers():
    p1 = providers.Provider()
    p2 = providers.Provider()

    container = ContainerA(p11=p1, p12=p2)

    assert container.p11.last_overriding is p1
    assert container.p12.last_overriding is p2


def test_init_with_overridden_dependency():
    # Bug: https://github.com/ets-labs/python-dependency-injector/issues/198
    class _Container(containers.DeclarativeContainer):
        p1 = providers.Dependency(instance_of=int)

        p2 = providers.Dependency(object)
        p2.override(providers.Factory(dict, p1=p1))

    container = _Container(p1=1)

    assert container.p2() == {"p1": 1}
    assert container.p2.last_overriding.kwargs["p1"] is container.p1
    assert container.p2.last_overriding.kwargs["p1"] is not _Container.p1
    assert _Container.p2.last_overriding.kwargs["p1"] is _Container.p1


def test_init_with_chained_dependency():
    # Bug: https://github.com/ets-labs/python-dependency-injector/issues/200
    class _Container(containers.DeclarativeContainer):
        p1 = providers.Dependency(instance_of=int)
        p2 = providers.Factory(p1)

    container = _Container(p1=1)

    assert container.p2() == 1
    assert container.p2.cls is container.p1
    assert _Container.p2.cls is _Container.p1
    assert container.p2.cls is not _Container.p1


def test_init_with_dependency_delegation():
    # Bug: https://github.com/ets-labs/python-dependency-injector/issues/235
    A = collections.namedtuple("A", [])
    B = collections.namedtuple("B", ["fa"])
    C = collections.namedtuple("B", ["a"])

    class Services(containers.DeclarativeContainer):
        a = providers.Dependency()
        c = providers.Factory(C, a=a)
        b = providers.Factory(B, fa=a.provider)

    a = providers.Factory(A)
    assert isinstance(Services(a=a).c().a, A)  # OK
    Services(a=a).b().fa()


def test_init_with_grand_child_provider():
    # Bug: https://github.com/ets-labs/python-dependency-injector/issues/350
    provider = providers.Provider()
    container = ContainerC(p11=provider)

    assert isinstance(container.p11, providers.Provider)
    assert isinstance(container.p12, providers.Provider)
    assert isinstance(container.p21, providers.Provider)
    assert isinstance(container.p22, providers.Provider)
    assert isinstance(container.p31, providers.Provider)
    assert isinstance(container.p32, providers.Provider)
    assert container.p11.last_overriding is provider


def test_parent_set_in__new__():
    class Container(containers.DeclarativeContainer):
        dependency = providers.Dependency()
        dependencies_container = providers.DependenciesContainer()
        container = providers.Container(ContainerA)

    assert Container.dependency.parent is Container
    assert Container.dependencies_container.parent is Container
    assert Container.container.parent is Container


def test_parent_set_in__setattr__():
    class Container(containers.DeclarativeContainer):
        pass

    Container.dependency = providers.Dependency()
    Container.dependencies_container = providers.DependenciesContainer()
    Container.container = providers.Container(ContainerA)

    assert Container.dependency.parent is Container
    assert Container.dependencies_container.parent is Container
    assert Container.container.parent is Container


def test_resolve_provider_name():
    assert ContainerA.resolve_provider_name(ContainerA.p11) == "p11"


def test_resolve_provider_name_no_provider():
    with raises(errors.Error):
        ContainerA.resolve_provider_name(providers.Provider())


def test_child_dependency_parent_name():
    class Container(containers.DeclarativeContainer):
        dependency = providers.Dependency()

    with raises(errors.Error, match="Dependency \"Container.dependency\" is not defined"):
        Container.dependency()


def test_child_dependencies_container_parent_name():
    class Container(containers.DeclarativeContainer):
        dependencies_container = providers.DependenciesContainer()

    with raises(errors.Error, match="Dependency \"Container.dependencies_container.dependency\" is not defined"):
        Container.dependencies_container.dependency()


def test_child_container_parent_name():
    class ChildContainer(containers.DeclarativeContainer):
        dependency = providers.Dependency()

    class Container(containers.DeclarativeContainer):
        child_container = providers.Container(ChildContainer)

    with raises(errors.Error, match="Dependency \"Container.child_container.dependency\" is not defined"):
        Container.child_container.dependency()



"""Delegated singleton provider tests."""

from dependency_injector import providers
from pytest import fixture

from .common import Example


PROVIDER_CLASSES = [
    providers.DelegatedSingleton,
    providers.DelegatedThreadLocalSingleton,
    providers.DelegatedThreadSafeSingleton,
]


@fixture(params=PROVIDER_CLASSES)
def singleton_cls(request):
    return request.param


@fixture
def provider(singleton_cls):
    return singleton_cls(Example)


def test_is_delegated_provider(provider):
    assert providers.is_delegated(provider) is True



"""Resources module."""

import abc
from typing import TypeVar, Generic, Optional


T = TypeVar("T")


class Resource(Generic[T], metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def init(self, *args, **kwargs) -> Optional[T]:
        ...

    def shutdown(self, resource: Optional[T]) -> None:
        ...


class AsyncResource(Generic[T], metaclass=abc.ABCMeta):

    @abc.abstractmethod
    async def init(self, *args, **kwargs) -> Optional[T]:
        ...

    async def shutdown(self, resource: Optional[T]) -> None:
        ...



"""Configuration.from_json() tests."""

from dependency_injector import errors
from pytest import mark, raises


def test(config, json_config_file_1):
    config.from_json(json_config_file_1)

    assert config() == {"section1": {"value1": 1}, "section2": {"value2": 2}}
    assert config.section1() == {"value1": 1}
    assert config.section1.value1() == 1
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2


def test_merge(config, json_config_file_1, json_config_file_2):
    config.from_json(json_config_file_1)
    config.from_json(json_config_file_2)

    assert config() == {
        "section1": {
            "value1": 11,
            "value11": 11,
        },
        "section2": {
            "value2": 2,
        },
        "section3": {
            "value3": 3,
        },
    }
    assert config.section1() == {"value1": 11, "value11": 11}
    assert config.section1.value1() == 11
    assert config.section1.value11() == 11
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2
    assert config.section3() == {"value3": 3}
    assert config.section3.value3() == 3


def test_file_does_not_exist(config):
    config.from_json("./does_not_exist.json")
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_file_does_not_exist_strict_mode(config):
    with raises(IOError):
        config.from_json("./does_not_exist.json")


def test_option_file_does_not_exist(config):
    config.option.from_json("./does_not_exist.json")
    assert config.option() is None


@mark.parametrize("config_type", ["strict"])
def test_option_file_does_not_exist_strict_mode(config):
    with raises(IOError):
        config.option.from_json("./does_not_exist.json")


def test_required_file_does_not_exist(config):
    with raises(IOError):
        config.from_json("./does_not_exist.json", required=True)


def test_required_option_file_does_not_exist(config):
    with raises(IOError):
        config.option.from_json("./does_not_exist.json", required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_file_does_not_exist_strict_mode(config):
    config.from_json("./does_not_exist.json", required=False)
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_not_required_option_file_does_not_exist_strict_mode(config):
    config.option.from_json("./does_not_exist.json", required=False)
    with raises(errors.Error):
        config.option()



"""Common test artifacts."""

import asyncio
import random

from dependency_injector import containers, providers


RESOURCE1 = object()
RESOURCE2 = object()


async def init_resource(resource):
    await asyncio.sleep(random.randint(1, 10) / 1000)
    yield resource
    await asyncio.sleep(random.randint(1, 10) / 1000)


class Client:
    def __init__(self, resource1: object, resource2: object) -> None:
        self.resource1 = resource1
        self.resource2 = resource2


class Service:
    def __init__(self, client: Client) -> None:
        self.client = client


class BaseContainer(containers.DeclarativeContainer):
    resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))
    resource2 = providers.Resource(init_resource, providers.Object(RESOURCE2))


class Container(BaseContainer):
    client = providers.Factory(
        Client,
        resource1=BaseContainer.resource1,
        resource2=BaseContainer.resource2,
    )

    service = providers.Factory(
        Service,
        client=client,
    )



from dependency_injector import containers, providers
from dependency_injector.wiring import inject, Provide, Closing


class Service:
    init_counter: int = 0
    shutdown_counter: int = 0

    @classmethod
    def reset_counter(cls):
        cls.init_counter = 0
        cls.shutdown_counter = 0

    @classmethod
    def init(cls):
        cls.init_counter += 1

    @classmethod
    def shutdown(cls):
        cls.shutdown_counter += 1


def init_service():
    service = Service()
    service.init()
    yield service
    service.shutdown()


class Container(containers.DeclarativeContainer):

    service = providers.Resource(init_service)


@inject
def test_function(service: Service = Closing[Provide[Container.service]]):
    return service



"""Schema module."""

import builtins
import importlib
from typing import Dict, Any, Type, Optional

from . import containers, providers


ContainerSchema = Dict[Any, Any]
ProviderSchema = Dict[Any, Any]


class SchemaProcessorV1:

    def __init__(self, schema: ContainerSchema) -> None:
        self._schema = schema
        self._container = containers.DynamicContainer()

    def process(self):
        """Process schema."""
        self._create_providers(self._schema["container"])
        self._setup_injections(self._schema["container"])

    def get_providers(self):
        """Return providers."""
        return self._container.providers

    def _create_providers(
            self,
            provider_schema: ProviderSchema,
            container: Optional[containers.Container] = None,
    ) -> None:
        if container is None:
            container = self._container
        for provider_name, data in provider_schema.items():
            provider = None

            if "provider" in data:
                provider_type = _get_provider_cls(data["provider"])
                args = []

                # provides = data.get("provides")
                # if provides:
                #     provides = _import_string(provides)
                #     if provides:
                #         args.append(provides)

                provider = provider_type(*args)

            if provider is None:
                provider = providers.Container(containers.DynamicContainer)

            container.set_provider(provider_name, provider)

            if isinstance(provider, providers.Container):
                self._create_providers(provider_schema=data, container=provider)

    def _setup_injections(  # noqa: C901
            self,
            provider_schema: ProviderSchema,
            container: Optional[containers.Container] = None,
    ) -> None:
        if container is None:
            container = self._container

        for provider_name, data in provider_schema.items():
            provider = getattr(container, provider_name)
            args = []
            kwargs = {}

            provides = data.get("provides")
            if provides:
                if isinstance(provides, str) and provides.startswith("container."):
                    provides = self._resolve_provider(provides[len("container."):])
                else:
                    provides = _import_string(provides)
                provider.set_provides(provides)

            arg_injections = data.get("args")
            if arg_injections:
                for arg in arg_injections:
                    injection = None

                    if isinstance(arg, str) and arg.startswith("container."):
                        injection = self._resolve_provider(arg[len("container."):])

                    # TODO: refactoring
                    if isinstance(arg, dict):
                        provider_args = []
                        provider_type = _get_provider_cls(arg.get("provider"))
                        provides = arg.get("provides")
                        if provides:
                            if isinstance(provides, str) and provides.startswith("container."):
                                provides = self._resolve_provider(provides[len("container."):])
                            else:
                                provides = _import_string(provides)
                            provider_args.append(provides)
                        for provider_arg in arg.get("args", []):
                            if isinstance(provider_arg, str) \
                                    and provider_arg.startswith("container."):
                                provider_args.append(
                                    self._resolve_provider(provider_arg[len("container."):]),
                                )
                        injection = provider_type(*provider_args)

                    if not injection:
                        injection = arg

                    args.append(injection)
            if args:
                provider.add_args(*args)

            kwarg_injections = data.get("kwargs")
            if kwarg_injections:
                for name, arg in kwarg_injections.items():
                    injection = None

                    if isinstance(arg, str) and arg.startswith("container."):
                        injection = self._resolve_provider(arg[len("container."):])

                    # TODO: refactoring
                    if isinstance(arg, dict):
                        provider_args = []
                        provider_type = _get_provider_cls(arg.get("provider"))
                        provides = arg.get("provides")
                        if provides:
                            if isinstance(provides, str) and provides.startswith("container."):
                                provides = self._resolve_provider(provides[len("container."):])
                            else:
                                provides = _import_string(provides)
                            provider_args.append(provides)
                        for provider_arg in arg.get("args", []):
                            if isinstance(provider_arg, str) \
                                    and provider_arg.startswith("container."):
                                provider_args.append(
                                    self._resolve_provider(provider_arg[len("container."):]),
                                )
                        injection = provider_type(*provider_args)

                    if not injection:
                        injection = arg

                    kwargs[name] = injection
            if kwargs:
                provider.add_kwargs(**kwargs)

            if isinstance(provider, providers.Container):
                self._setup_injections(provider_schema=data, container=provider)

    def _resolve_provider(self, name: str) -> Optional[providers.Provider]:
        segments = name.split(".")
        try:
            provider = getattr(self._container, segments[0])
        except AttributeError:
            return None

        for segment in segments[1:]:
            parentheses = ""
            if "(" in segment and ")" in segment:
                parentheses = segment[segment.find("("):segment.rfind(")")+1]
                segment = segment.replace(parentheses, "")

            try:
                provider = getattr(provider, segment)
            except AttributeError:
                # TODO
                return None

            if parentheses:
                # TODO
                provider = provider()

        return provider


def build_schema(schema: ContainerSchema) -> Dict[str, providers.Provider]:
    """Build provider schema."""
    schema_processor = SchemaProcessorV1(schema)
    schema_processor.process()
    return schema_processor.get_providers()


def _get_provider_cls(provider_cls_name: str) -> Type[providers.Provider]:
    std_provider_type = _fetch_provider_cls_from_std(provider_cls_name)
    if std_provider_type:
        return std_provider_type

    custom_provider_type = _import_provider_cls(provider_cls_name)
    if custom_provider_type:
        return custom_provider_type

    raise SchemaError(f"Undefined provider class \"{provider_cls_name}\"")


def _fetch_provider_cls_from_std(provider_cls_name: str) -> Optional[Type[providers.Provider]]:
    return getattr(providers, provider_cls_name, None)


def _import_provider_cls(provider_cls_name: str) -> Optional[Type[providers.Provider]]:
    try:
        cls = _import_string(provider_cls_name)
    except (ImportError, ValueError) as exception:
        raise SchemaError(f"Can not import provider \"{provider_cls_name}\"") from exception
    except AttributeError:
        return None
    else:
        if isinstance(cls, type) and not issubclass(cls, providers.Provider):
            raise SchemaError(f"Provider class \"{cls}\" is not a subclass of providers base class")
        return cls


def _import_string(string_name: str) -> Optional[object]:
    segments = string_name.split(".")

    if len(segments) == 1:
        member = getattr(builtins, segments[0], None)
        if member:
            return member

    module_name = ".".join(segments[:-1])
    if not module_name:
        return None

    member = segments[-1]
    module = importlib.import_module(module_name)
    return getattr(module, member, None)


class SchemaError(Exception):
    """Schema-related error."""



"""Configuration.from_pydantic() tests."""

from pydantic import BaseModel

try:
    from pydantic_settings import (
        BaseSettings,  # type: ignore[import-not-found,unused-ignore]
    )
except ImportError:
    try:
        from pydantic import BaseSettings  # type: ignore[no-redef,unused-ignore]
    except ImportError:

        class BaseSettings:  # type: ignore[no-redef]
            """No-op fallback"""


from pytest import fixture, mark, raises

from dependency_injector import errors, providers

pytestmark = mark.pydantic


class Section11(BaseModel):
    value1: int = 1


class Section12(BaseModel):
    value2: int = 2


class Settings1(BaseSettings):
    section1: Section11 = Section11()
    section2: Section12 = Section12()


class Section21(BaseModel):
    value1: int = 11
    value11: int = 11


class Section3(BaseModel):
    value3: int = 3


class Settings2(BaseSettings):
    section1: Section21 = Section21()
    section3: Section3 = Section3()


@fixture
def no_pydantic_module_installed():
    has_pydantic_settings = providers.has_pydantic_settings
    providers.has_pydantic_settings = False
    yield
    providers.has_pydantic_settings = has_pydantic_settings


def test(config):
    config.from_pydantic(Settings1())

    assert config() == {"section1": {"value1": 1}, "section2": {"value2": 2}}
    assert config.section1() == {"value1": 1}
    assert config.section1.value1() == 1
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2


def test_kwarg(config):
    config.from_pydantic(Settings1(), exclude={"section2"})

    assert config() == {"section1": {"value1": 1}}
    assert config.section1() == {"value1": 1}
    assert config.section1.value1() == 1


def test_merge(config):
    config.from_pydantic(Settings1())
    config.from_pydantic(Settings2())

    assert config() == {
        "section1": {
            "value1": 11,
            "value11": 11,
        },
        "section2": {
            "value2": 2,
        },
        "section3": {
            "value3": 3,
        },
    }
    assert config.section1() == {"value1": 11, "value11": 11}
    assert config.section1.value1() == 11
    assert config.section1.value11() == 11
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2
    assert config.section3() == {"value3": 3}
    assert config.section3.value3() == 3


def test_empty_settings(config):
    config.from_pydantic(BaseSettings())
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_empty_settings_strict_mode(config):
    with raises(ValueError):
        config.from_pydantic(BaseSettings())


def test_option_empty_settings(config):
    config.option.from_pydantic(BaseSettings())
    assert config.option() == {}


@mark.parametrize("config_type", ["strict"])
def test_option_empty_settings_strict_mode(config):
    with raises(ValueError):
        config.option.from_pydantic(BaseSettings())


def test_required_empty_settings(config):
    with raises(ValueError):
        config.from_pydantic(BaseSettings(), required=True)


def test_required_option_empty_settings(config):
    with raises(ValueError):
        config.option.from_pydantic(BaseSettings(), required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_empty_settings_strict_mode(config):
    config.from_pydantic(BaseSettings(), required=False)
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_not_required_option_empty_settings_strict_mode(config):
    config.option.from_pydantic(BaseSettings(), required=False)
    assert config.option() == {}
    assert config() == {"option": {}}


def test_not_instance_of_settings(config):
    with raises(
        errors.Error,
        match=(
            r"Unable to recognize settings instance, expect \"pydantic(?:_settings)?\.BaseSettings\", "
            r"got {0} instead".format({})
        ),
    ):
        config.from_pydantic({})


def test_option_not_instance_of_settings(config):
    with raises(
        errors.Error,
        match=(
            r"Unable to recognize settings instance, expect \"pydantic(?:_settings)?\.BaseSettings\", "
            "got {0} instead".format({})
        ),
    ):
        config.option.from_pydantic({})


def test_subclass_instead_of_instance(config):
    with raises(errors.Error) as error:
        config.from_pydantic(Settings1)
    assert error.value.args[0] == (
        "Got settings class, but expect instance: "
        "instead \"Settings1\" use \"Settings1()\""
    )


def test_option_subclass_instead_of_instance(config):
    with raises(errors.Error) as error:
        config.option.from_pydantic(Settings1)
    assert error.value.args[0] == (
        "Got settings class, but expect instance: "
        "instead \"Settings1\" use \"Settings1()\""
    )


@mark.usefixtures("no_pydantic_module_installed")
def test_no_pydantic_installed(config):
    with raises(
        errors.Error,
        match=(
            r"Unable to load pydantic configuration - pydantic(?:_settings)? is not installed\. "
            r"Install pydantic or install Dependency Injector with pydantic extras: "
            r"\"pip install dependency-injector\[pydantic2?\]\""
        ),
    ):
        config.from_pydantic(Settings1())


@mark.usefixtures("no_pydantic_module_installed")
def test_option_no_pydantic_installed(config):
    with raises(
        errors.Error,
        match=(
            r"Unable to load pydantic configuration - pydantic(?:_settings)? is not installed\. "
            r"Install pydantic or install Dependency Injector with pydantic extras: "
            r"\"pip install dependency-injector\[pydantic2?\]\""
        ),
    ):
        config.option.from_pydantic(Settings1())



from typing import Awaitable as _Awaitable

from dependency_injector import providers


class Application(providers.Singleton): ...


class Extension(providers.Singleton): ...


class Middleware(providers.DelegatedCallable): ...


class MiddlewareFactory(providers.Factory): ...


class View(providers.Callable):
    def as_view(self) -> _Awaitable: ...


class ClassBasedView(providers.Factory):
    def as_view(self) -> _Awaitable: ...



"""Tests for provider overriding in async mode."""

from dependency_injector import providers
from pytest import mark


@mark.asyncio
async def test_provider():
    dependency = object()

    async def _get_dependency_async():
        return dependency

    def _get_dependency_sync():
        return dependency

    provider = providers.Provider()

    provider.override(providers.Callable(_get_dependency_async))
    dependency1 = await provider()

    provider.override(providers.Callable(_get_dependency_sync))
    dependency2 = await provider()

    assert dependency1 is dependency
    assert dependency2 is dependency


@mark.asyncio
async def test_callable():
    dependency = object()

    async def _get_dependency_async():
        return dependency

    def _get_dependency_sync():
        return dependency

    provider = providers.Callable(_get_dependency_async)
    dependency1 = await provider()

    provider.override(providers.Callable(_get_dependency_sync))
    dependency2 = await provider()

    assert dependency1 is dependency
    assert dependency2 is dependency


@mark.asyncio
async def test_factory():
    dependency = object()

    async def _get_dependency_async():
        return dependency

    def _get_dependency_sync():
        return dependency

    provider = providers.Factory(_get_dependency_async)
    dependency1 = await provider()

    provider.override(providers.Callable(_get_dependency_sync))
    dependency2 = await provider()

    assert dependency1 is dependency
    assert dependency2 is dependency


@mark.asyncio
async def test_async_mode_enabling():
    dependency = object()

    async def _get_dependency_async():
        return dependency

    provider = providers.Callable(_get_dependency_async)
    assert provider.is_async_mode_undefined() is True

    await provider()

    assert provider.is_async_mode_enabled() is True


@mark.asyncio
async def test_async_mode_disabling():
    dependency = object()

    def _get_dependency():
        return dependency

    provider = providers.Callable(_get_dependency)
    assert provider.is_async_mode_undefined() is True

    provider()

    assert provider.is_async_mode_disabled() is True


@mark.asyncio
async def test_async_mode_enabling_on_overriding():
    dependency = object()

    async def _get_dependency_async():
        return dependency

    provider = providers.Provider()
    provider.override(providers.Callable(_get_dependency_async))
    assert provider.is_async_mode_undefined() is True

    await provider()

    assert provider.is_async_mode_enabled() is True


def test_async_mode_disabling_on_overriding():
    dependency = object()

    def _get_dependency():
        return dependency

    provider = providers.Provider()
    provider.override(providers.Callable(_get_dependency))
    assert provider.is_async_mode_undefined() is True

    provider()

    assert provider.is_async_mode_disabled() is True



"""MethodCaller provider traversal tests."""

from dependency_injector import providers


def test_traverse():
    provider1 = providers.Provider()
    provided = provider1.provided
    method = provided.method
    provider = method.call()

    all_providers = list(provider.traverse())

    assert len(all_providers) == 3
    assert provider1 in all_providers
    assert provided in all_providers
    assert method in all_providers


def test_traverse_args():
    provider1 = providers.Provider()
    provided = provider1.provided
    method = provided.method
    provider2 = providers.Provider()
    provider = method.call("foo", provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 4
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert provided in all_providers
    assert method in all_providers


def test_traverse_kwargs():
    provider1 = providers.Provider()
    provided = provider1.provided
    method = provided.method
    provider2 = providers.Provider()
    provider = method.call(foo="foo", bar=provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 4
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert provided in all_providers
    assert method in all_providers


def test_traverse_overridden():
    provider1 = providers.Provider()
    provided = provider1.provided
    method = provided.method
    provider2 = providers.Provider()

    provider = method.call()
    provider.override(provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 4
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert provided in all_providers
    assert method in all_providers



"""Test module for wiring with invalid type of marker for attribute injection."""

from dependency_injector.wiring import Closing

from .container import Container


service = Closing[Container.service]



import json

# Runtime import to avoid syntax errors in samples on Python < 3.5 and reach top-dir
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        "../",
    )),
)
_SAMPLES_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        "../samples/",
    )),
)
import sys
sys.path.append(_TOP_DIR)
sys.path.append(_SAMPLES_DIR)

from wiringflask import web


def test_wiring_with_flask():
    client = web.app.test_client()

    with web.app.app_context():
        response = client.get("/")

    assert response.status_code == 200
    assert json.loads(response.data) == {"result": "OK"}



"""Tests for wiring causes no issues with queue.Queue from std lib."""

from pytest import fixture

from samples.wiring import queuemodule
from samples.wiring.container import Container


@fixture
def container():
    container = Container()
    yield container
    container.unwire()


def test_wire_queue(container: Container):
    # See: https://github.com/ets-labs/python-dependency-injector/issues/362
    # Should not raise exception
    try:
        container.wire(modules=[queuemodule])
    except:
        raise



"""Factory delegate provider tests."""

from dependency_injector import providers, errors
from pytest import fixture, raises


@fixture
def factory():
    return providers.Factory(object)


@fixture
def delegate(factory):
    return providers.FactoryDelegate(factory)


def test_is_delegate(delegate):
    assert isinstance(delegate, providers.Delegate)


def test_init_with_not_factory():
    with raises(errors.Error):
        providers.FactoryDelegate(providers.Object(object()))



"""Flask extension tests."""

from dependency_injector import containers
from dependency_injector.ext import flask
from flask import Flask, url_for
from flask.views import MethodView
from pytest import fixture


def index():
    return "Hello World!"


def test():
    return "Test!"


class Test(MethodView):
    def get(self):
        return "Test class-based!"


class ApplicationContainer(containers.DeclarativeContainer):

    app = flask.Application(Flask, __name__)

    index_view = flask.View(index)
    test_view = flask.View(test)
    test_class_view = flask.ClassBasedView(Test)


@fixture
def app():
    container = ApplicationContainer()
    app = container.app()
    app.container = container
    app.config["SERVER_NAME"] = "test-server.com"
    app.add_url_rule("/", view_func=container.index_view.as_view())
    app.add_url_rule("/test", "test-test", view_func=container.test_view.as_view())
    app.add_url_rule("/test-class", view_func=container.test_class_view.as_view("test-class"))
    return app


@fixture
def client(app):
    with app.test_client() as client:
        yield client


def test_index(client):
    response = client.get("/")

    assert response.status_code == 200
    assert response.data == b"Hello World!"


def test_test(client):
    response = client.get("/test")

    assert response.status_code == 200
    assert response.data == b"Test!"


def test_test_class_based(client):
    response = client.get("/test-class")

    assert response.status_code == 200
    assert response.data == b"Test class-based!"


def test_endpoints(app):
    with app.app_context():
        assert url_for("index") == "http://test-server.com/"
        assert url_for("test-test") == "http://test-server.com/test"
        assert url_for("test-class") == "http://test-server.com/test-class"



"""Tests for string module and package names."""

from pytest import fixture

from samples.wiring import module
from samples.wiring.service import Service
from samples.wiring.container import Container
from samples.wiring.wire_relative_string_names import wire_with_relative_string_names


@fixture
def container():
    container = Container()
    yield container
    container.unwire()


def test_absolute_names(container: Container):
    container.wire(
        modules=["samples.wiring.module"],
        packages=["samples.wiring.package"],
    )

    service = module.test_function()
    assert isinstance(service, Service)

    from samples.wiring.package.subpackage.submodule import test_function
    service = test_function()
    assert isinstance(service, Service)


def test_relative_names_with_explicit_package(container: Container):
    container.wire(
        modules=[".module"],
        packages=[".package"],
        from_package="samples.wiring",
    )

    service = module.test_function()
    assert isinstance(service, Service)

    from samples.wiring.package.subpackage.submodule import test_function
    service = test_function()
    assert isinstance(service, Service)


def test_relative_names_with_auto_package(container: Container):
    wire_with_relative_string_names(container)

    service = module.test_function()
    assert isinstance(service, Service)

    from samples.wiring.package.subpackage.submodule import test_function
    service = test_function()
    assert isinstance(service, Service)





"""Async injection tests."""

from pytest import fixture, mark

from samples.wiringstringids import asyncinjections


@fixture(autouse=True)
def container():
    container = asyncinjections.Container()
    container.wire(modules=[asyncinjections])
    yield container
    container.unwire()


@fixture(autouse=True)
def reset_counters():
    asyncinjections.resource1.reset_counters()
    asyncinjections.resource2.reset_counters()


@mark.asyncio
async def test_async_injections():
    resource1, resource2 = await asyncinjections.async_injection()

    assert resource1 is asyncinjections.resource1
    assert asyncinjections.resource1.init_counter == 1
    assert asyncinjections.resource1.shutdown_counter == 0

    assert resource2 is asyncinjections.resource2
    assert asyncinjections.resource2.init_counter == 1
    assert asyncinjections.resource2.shutdown_counter == 0


@mark.asyncio
async def test_async_injections_with_closing():
    resource1, resource2 = await asyncinjections.async_injection_with_closing()

    assert resource1 is asyncinjections.resource1
    assert asyncinjections.resource1.init_counter == 1
    assert asyncinjections.resource1.shutdown_counter == 1

    assert resource2 is asyncinjections.resource2
    assert asyncinjections.resource2.init_counter == 1
    assert asyncinjections.resource2.shutdown_counter == 1

    resource1, resource2 = await asyncinjections.async_injection_with_closing()

    assert resource1 is asyncinjections.resource1
    assert asyncinjections.resource1.init_counter == 2
    assert asyncinjections.resource1.shutdown_counter == 2

    assert resource2 is asyncinjections.resource2
    assert asyncinjections.resource2.init_counter == 2
    assert asyncinjections.resource2.shutdown_counter == 2



"""AbstractCoroutine provider tests."""

import asyncio
import sys

from dependency_injector import providers, errors
from pytest import mark, raises

from .common import example


def test_inheritance():
    assert isinstance(providers.AbstractCoroutine(example), providers.Coroutine)


@mark.skipif(sys.version_info >= (3, 11), reason="asyncio.coroutine removed in Python 3.11")
@mark.asyncio
@mark.filterwarnings("ignore")
async def test_call_overridden_by_coroutine():
    @asyncio.coroutine
    def abstract_example():
        raise RuntimeError("Should not be raised")

    provider = providers.AbstractCoroutine(abstract_example)
    provider.override(providers.Coroutine(example))

    result = await provider(1, 2, 3, 4)
    assert result == (1, 2, 3, 4)


@mark.skipif(sys.version_info >= (3, 11), reason="asyncio.coroutine removed in Python 3.11")
@mark.asyncio
@mark.filterwarnings("ignore")
async def test_call_overridden_by_delegated_coroutine():
    @asyncio.coroutine
    def abstract_example():
        raise RuntimeError("Should not be raised")

    provider = providers.AbstractCoroutine(abstract_example)
    provider.override(providers.DelegatedCoroutine(example))

    result = await provider(1, 2, 3, 4)
    assert result == (1, 2, 3, 4)


def test_call_not_overridden():
    provider = providers.AbstractCoroutine(example)
    with raises(errors.Error):
        provider(1, 2, 3, 4)


def test_override_by_not_coroutine():
    provider = providers.AbstractCoroutine(example)
    with raises(errors.Error):
        provider.override(providers.Factory(object))


def test_provide_not_implemented():
    provider = providers.AbstractCoroutine(example)
    with raises(NotImplementedError):
        provider._provide((1, 2, 3, 4), dict())


def test_repr():
    provider = providers.AbstractCoroutine(example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "AbstractCoroutine({0}) at {1}>".format(repr(example), hex(id(provider)))
    )



"""Singleton provider tests."""
import decimal
import sys

from dependency_injector import providers, errors
from pytest import fixture, raises, mark

from .common import Example


PROVIDER_CLASSES = [
    providers.Singleton,
    providers.DelegatedSingleton,
    providers.ThreadLocalSingleton,
    providers.DelegatedThreadLocalSingleton,
    providers.ThreadSafeSingleton,
    providers.DelegatedThreadSafeSingleton,
]
if sys.version_info >= (3, 5):
    PROVIDER_CLASSES.append(providers.ContextLocalSingleton)


@fixture(params=PROVIDER_CLASSES)
def singleton_cls(request):
    return request.param


@fixture
def provider(singleton_cls):
    return singleton_cls(Example)


def test_is_provider(provider):
    assert providers.is_provider(provider) is True


def test_init_with_not_callable(singleton_cls):
    with raises(errors.Error):
        singleton_cls(123)


def test_init_optional_provides(provider):
    provider.set_provides(object)
    assert provider.provides is object
    assert isinstance(provider(), object)


def test_set_provides_returns_self(provider):
    assert provider.set_provides(object) is provider


@mark.parametrize(
    "str_name,cls",
    [
        ("dependency_injector.providers.Factory", providers.Factory),
        ("decimal.Decimal", decimal.Decimal),
        ("list", list),
        (".common.Example", Example),
        ("test_is_provider", test_is_provider),
    ],
)
def test_set_provides_string_imports(str_name, cls):
    assert providers.Singleton(str_name).provides is cls


def test_init_with_valid_provided_type(singleton_cls):
    class ExampleProvider(singleton_cls):
        provided_type = Example

    example_provider = ExampleProvider(Example, 1, 2)
    assert isinstance(example_provider(), Example)


def test_init_with_valid_provided_subtype(singleton_cls):
    class ExampleProvider(singleton_cls):
        provided_type = Example

    class NewExample(Example):
        pass

    example_provider = ExampleProvider(NewExample, 1, 2)
    assert isinstance(example_provider(), NewExample)


def test_init_with_invalid_provided_type(singleton_cls):
    class ExampleProvider(singleton_cls):
        provided_type = Example

    with raises(errors.Error):
        ExampleProvider(list)


def test_provided_instance_provider(provider):
    assert isinstance(provider.provided, providers.ProvidedInstance)


def test_call(provider):
    instance1 = provider()
    instance2 = provider()

    assert instance1 is instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_init_positional_args(singleton_cls):
    provider = singleton_cls(Example, "i1", "i2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.init_arg1 == "i1"
    assert instance1.init_arg2 == "i2"

    assert instance2.init_arg1 == "i1"
    assert instance2.init_arg2 == "i2"

    assert instance1 is instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_init_keyword_args(singleton_cls):
    provider = singleton_cls(Example, init_arg1="i1", init_arg2="i2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.init_arg1 == "i1"
    assert instance1.init_arg2 == "i2"

    assert instance2.init_arg1 == "i1"
    assert instance2.init_arg2 == "i2"

    assert instance1 is instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_init_positional_and_keyword_args(singleton_cls):
    provider = singleton_cls(Example, "i1", init_arg2="i2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.init_arg1 == "i1"
    assert instance1.init_arg2 == "i2"

    assert instance2.init_arg1 == "i1"
    assert instance2.init_arg2 == "i2"

    assert instance1 is instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_attributes(provider):
    provider.add_attributes(attribute1="a1", attribute2="a2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.attribute1 == "a1"
    assert instance1.attribute2 == "a2"

    assert instance2.attribute1 == "a1"
    assert instance2.attribute2 == "a2"

    assert instance1 is instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_context_args(provider):
    instance = provider(11, 22)

    assert instance.init_arg1 == 11
    assert instance.init_arg2 == 22


def test_call_with_context_kwargs(singleton_cls):
    provider = singleton_cls(Example, init_arg1=1)

    instance1 = provider(init_arg2=22)
    assert instance1.init_arg1 == 1
    assert instance1.init_arg2 == 22

    # Instance is created earlier
    instance1 = provider(init_arg1=11, init_arg2=22)
    assert instance1.init_arg1 == 1
    assert instance1.init_arg2 == 22


def test_call_with_context_args_and_kwargs(singleton_cls):
    provider = singleton_cls(Example, 11)

    instance = provider(22, init_arg3=33, init_arg4=44)

    assert instance.init_arg1 == 11
    assert instance.init_arg2 == 22
    assert instance.init_arg3 == 33
    assert instance.init_arg4 == 44


def test_fluent_interface(singleton_cls):
    provider = singleton_cls(Example) \
        .add_args(1, 2) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .add_attributes(attribute1=5, attribute2=6)

    instance = provider()

    assert instance.init_arg1 == 1
    assert instance.init_arg2 == 2
    assert instance.init_arg3 == 3
    assert instance.init_arg4 == 4
    assert instance.attribute1 == 5
    assert instance.attribute2 == 6


def test_set_args(singleton_cls):
    provider = singleton_cls(Example) \
        .add_args(1, 2) \
        .set_args(3, 4)
    assert provider.args == (3, 4)


def test_set_kwargs(singleton_cls):
    provider = singleton_cls(Example) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .set_kwargs(init_arg3=4, init_arg4=5)
    assert provider.kwargs == dict(init_arg3=4, init_arg4=5)


def test_set_attributes(singleton_cls):
    provider = singleton_cls(Example) \
        .add_attributes(attribute1=5, attribute2=6) \
        .set_attributes(attribute1=6, attribute2=7)
    assert provider.attributes == dict(attribute1=6, attribute2=7)


def test_clear_args(singleton_cls):
    provider = singleton_cls(Example) \
        .add_args(1, 2) \
        .clear_args()
    assert provider.args == tuple()


def test_clear_kwargs(singleton_cls):
    provider = singleton_cls(Example) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .clear_kwargs()
    assert provider.kwargs == dict()


def test_clear_attributes(singleton_cls):
    provider = singleton_cls(Example) \
        .add_attributes(attribute1=5, attribute2=6) \
        .clear_attributes()
    assert provider.attributes == dict()


def test_call_overridden(singleton_cls):
    provider = singleton_cls(Example)
    overriding_provider1 = singleton_cls(dict)
    overriding_provider2 = singleton_cls(list)

    provider.override(overriding_provider1)
    provider.override(overriding_provider2)

    instance1 = provider()
    instance2 = provider()

    assert instance1 is instance2
    assert isinstance(instance1, list)
    assert isinstance(instance2, list)


def test_deepcopy(singleton_cls):
    provider = singleton_cls(Example)
    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert provider.cls is provider_copy.cls
    assert isinstance(provider, singleton_cls)


def test_deepcopy_from_memo(singleton_cls):
    provider = singleton_cls(Example)
    provider_copy_memo = singleton_cls(Example)

    provider_copy = providers.deepcopy(provider, memo={id(provider): provider_copy_memo})
    assert provider_copy is provider_copy_memo


def test_deepcopy_args(singleton_cls):
    provider = singleton_cls(Example)
    dependent_provider1 = singleton_cls(list)
    dependent_provider2 = singleton_cls(dict)

    provider.add_args(dependent_provider1, dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.args[0]
    dependent_provider_copy2 = provider_copy.args[1]

    assert provider.args != provider_copy.args

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_kwargs(singleton_cls):
    provider = singleton_cls(Example)
    dependent_provider1 = singleton_cls(list)
    dependent_provider2 = singleton_cls(dict)

    provider.add_kwargs(a1=dependent_provider1, a2=dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.kwargs["a1"]
    dependent_provider_copy2 = provider_copy.kwargs["a2"]

    assert provider.kwargs != provider_copy.kwargs

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_attributes(singleton_cls):
    provider = singleton_cls(Example)
    dependent_provider1 = singleton_cls(list)
    dependent_provider2 = singleton_cls(dict)

    provider.add_attributes(a1=dependent_provider1, a2=dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.attributes["a1"]
    dependent_provider_copy2 = provider_copy.attributes["a2"]

    assert provider.attributes != provider_copy.attributes

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_overridden(singleton_cls):
    provider = singleton_cls(Example)
    object_provider = providers.Object(object())

    provider.override(object_provider)

    provider_copy = providers.deepcopy(provider)
    object_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert provider.cls is provider_copy.cls
    assert isinstance(provider, singleton_cls)

    assert object_provider is not object_provider_copy
    assert isinstance(object_provider_copy, providers.Object)


def test_deepcopy_with_sys_streams(singleton_cls):
    provider = singleton_cls(Example)
    provider.add_args(sys.stdin)
    provider.add_kwargs(a2=sys.stdout)
    provider.add_attributes(a3=sys.stderr)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider_copy, singleton_cls)
    assert provider.args[0] is sys.stdin
    assert provider.kwargs["a2"] is sys.stdout
    assert provider.attributes["a3"] is sys.stderr


def test_reset(singleton_cls):
    provider = singleton_cls(object)

    instance1 = provider()
    assert isinstance(instance1, object)

    provider.reset()

    instance2 = provider()
    assert isinstance(instance2, object)
    assert instance1 is not instance2


def test_reset_clean(provider):
    instance1 = provider()

    provider.reset()
    provider.reset()

    instance2 = provider()
    assert instance1 is not instance2


def test_reset_with_singleton(singleton_cls):
    dependent_singleton = providers.Singleton(object)
    provider = singleton_cls(dict, dependency=dependent_singleton)

    dependent_instance = dependent_singleton()
    instance1 = provider()
    assert instance1["dependency"] is dependent_instance

    provider.reset()

    instance2 = provider()
    assert instance2["dependency"] is dependent_instance
    assert instance1 is not instance2


def test_reset_context_manager(provider):
    instance1 = provider()
    with provider.reset():
        instance2 = provider()
    instance3 = provider()
    assert len({instance1, instance2, instance3}) == 3


def test_reset_context_manager_as_attribute(provider):
    with provider.reset() as alias:
        pass
    assert provider is alias


def test_full_reset(singleton_cls):
    dependent_singleton = providers.Singleton(object)
    provider = singleton_cls(dict, dependency=dependent_singleton)

    dependent_instance1 = dependent_singleton()
    instance1 = provider()
    assert instance1["dependency"] is dependent_instance1

    provider.full_reset()

    dependent_instance2 = dependent_singleton()
    instance2 = provider()
    assert instance2["dependency"] is not dependent_instance1
    assert dependent_instance1 is not dependent_instance2
    assert instance1 is not instance2


def test_full_reset_context_manager(singleton_cls):
    class Item:
        def __init__(self, dependency):
            self.dependency = dependency

    dependent_singleton = providers.Singleton(object)
    singleton = singleton_cls(Item, dependency=dependent_singleton)

    instance1 = singleton()
    with singleton.full_reset():
        instance2 = singleton()
    instance3 = singleton()

    assert len({instance1, instance2, instance3}) == 3
    assert len({instance1.dependency, instance2.dependency, instance3.dependency}) == 3


def test_full_reset_context_manager_as_attribute(provider):
    with provider.full_reset() as alias:
        pass
    assert provider is alias


def test_repr(provider):
    assert repr(provider) == (
        "<dependency_injector.providers."
        "{0}({1}) at {2}>".format(provider.__class__.__name__, repr(Example), hex(id(provider)))
    )



"""Tests for container async resources."""

import asyncio

from dependency_injector import containers, providers
from pytest import mark, raises


@mark.asyncio
async def test_init_and_shutdown_ordering():
    """Test init and shutdown resources.

    Methods .init_resources() and .shutdown_resources() should respect resources dependencies.
    Initialization should first initialize resources without dependencies and then provide
    these resources to other resources. Resources shutdown should follow the same rule: first
    shutdown resources without initialized dependencies and then continue correspondingly
    until all resources are shutdown.
    """
    initialized_resources = []
    shutdown_resources = []

    async def _resource(name, delay, **_):
        await asyncio.sleep(delay)
        initialized_resources.append(name)

        yield name

        await asyncio.sleep(delay)
        shutdown_resources.append(name)

    class Container(containers.DeclarativeContainer):
        resource1 = providers.Resource(
            _resource,
            name="r1",
            delay=0.03,
        )
        resource2 = providers.Resource(
            _resource,
            name="r2",
            delay=0.02,
            r1=resource1,
        )
        resource3 = providers.Resource(
            _resource,
            name="r3",
            delay=0.01,
            r2=resource2,
        )

    container = Container()

    await container.init_resources()
    assert initialized_resources == ["r1", "r2", "r3"]
    assert shutdown_resources == []

    await container.shutdown_resources()
    assert initialized_resources == ["r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1"]

    await container.init_resources()
    assert initialized_resources == ["r1", "r2", "r3", "r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1"]

    await container.shutdown_resources()
    assert initialized_resources == ["r1", "r2", "r3", "r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1", "r3", "r2", "r1"]


@mark.asyncio
async def test_shutdown_circular_dependencies_breaker():
    async def _resource(name, **_):
        yield name

    class Container(containers.DeclarativeContainer):
        resource1 = providers.Resource(
            _resource,
            name="r1",
        )
        resource2 = providers.Resource(
            _resource,
            name="r2",
            r1=resource1,
        )
        resource3 = providers.Resource(
            _resource,
            name="r3",
            r2=resource2,
        )

    container = Container()
    await container.init_resources()

    # Create circular dependency after initialization (r3 -> r2 -> r1 -> r3 -> ...)
    container.resource1.add_kwargs(r3=container.resource3)

    with raises(RuntimeError, match="Unable to resolve resources shutdown order"):
        await container.shutdown_resources()


@mark.asyncio
async def test_shutdown_sync_and_async_ordering():
    initialized_resources = []
    shutdown_resources = []

    def _sync_resource(name, **_):
        initialized_resources.append(name)
        yield name
        shutdown_resources.append(name)

    async def _async_resource(name, **_):
        initialized_resources.append(name)
        yield name
        shutdown_resources.append(name)

    class Container(containers.DeclarativeContainer):
        resource1 = providers.Resource(
            _sync_resource,
            name="r1",
        )
        resource2 = providers.Resource(
            _sync_resource,
            name="r2",
            r1=resource1,
        )
        resource3 = providers.Resource(
            _async_resource,
            name="r3",
            r2=resource2,
        )

    container = Container()

    await container.init_resources()
    assert initialized_resources == ["r1", "r2", "r3"]
    assert shutdown_resources == []

    await container.shutdown_resources()
    assert initialized_resources == ["r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1"]

    await container.init_resources()
    assert initialized_resources == ["r1", "r2", "r3", "r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1"]

    await container.shutdown_resources()
    assert initialized_resources == ["r1", "r2", "r3", "r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1", "r3", "r2", "r1"]



"""Dependency provider tests."""

from dependency_injector import containers, providers, errors
from pytest import fixture, raises


@fixture
def provider():
    return providers.Dependency(instance_of=list)


def test_init_optional():
    list_provider = providers.List(1, 2, 3)
    provider = providers.Dependency()
    provider.set_instance_of(list)
    provider.set_default(list_provider)

    assert provider.instance_of is list
    assert provider.default is list_provider
    assert provider() == [1, 2, 3]


def test_set_instance_of_returns_self(provider):
    assert provider.set_instance_of(list) is provider


def test_set_default_returns_self(provider):
    assert provider.set_default(providers.Provider()) is provider


def test_init_with_not_class():
    with raises(TypeError):
        providers.Dependency(object())


def test_with_abc():
    try:
        import collections.abc as collections_abc
    except ImportError:
        import collections as collections_abc

    provider = providers.Dependency(collections_abc.Mapping)
    provider.provided_by(providers.Factory(dict))

    assert isinstance(provider(), collections_abc.Mapping)
    assert isinstance(provider(), dict)


def test_is_provider(provider):
    assert providers.is_provider(provider) is True


def test_provided_instance_provider(provider):
    assert isinstance(provider.provided, providers.ProvidedInstance)


def test_default():
    provider = providers.Dependency(instance_of=dict, default={"foo": "bar"})
    assert provider() == {"foo": "bar"}


def test_default_attribute():
    provider = providers.Dependency(instance_of=dict, default={"foo": "bar"})
    assert provider.default() == {"foo": "bar"}


def test_default_provider():
    provider = providers.Dependency(instance_of=dict, default=providers.Factory(dict, foo="bar"))
    assert provider.default() == {"foo": "bar"}


def test_default_attribute_provider():
    default = providers.Factory(dict, foo="bar")
    provider = providers.Dependency(instance_of=dict, default=default)

    assert provider.default() == {"foo": "bar"}
    assert provider.default is default


def test_default_with_empty_dict():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/550
    default = {}
    provider = providers.Dependency(instance_of=dict, default=default)
    assert provider() == default
    assert provider.default() == default


def test_default_with_empty_string():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/550
    default = ""
    provider = providers.Dependency(instance_of=str, default=default)
    assert provider() == default
    assert provider.default() == default


def test_is_defined(provider):
    assert provider.is_defined is False


def test_is_defined_when_overridden(provider):
    provider.override("value")
    assert provider.is_defined is True


def test_is_defined_with_default():
    provider = providers.Dependency(default="value")
    assert provider.is_defined is True


def test_call_overridden(provider):
    provider.provided_by(providers.Factory(list))
    assert isinstance(provider(), list)


def test_call_overridden_but_not_instance_of(provider):
    provider.provided_by(providers.Factory(dict))
    with raises(errors.Error):
        provider()


def test_call_undefined(provider):
    with raises(errors.Error, match="Dependency is not defined"):
        provider()


def test_call_undefined_error_message_with_container_instance_parent():
    class UserService:
        def __init__(self, database):
            self.database = database

    class Container(containers.DeclarativeContainer):
        database = providers.Dependency()

        user_service = providers.Factory(
            UserService,
            database=database,  # <---- missing dependency
        )

    container = Container()

    with raises(errors.Error, match="Dependency \"Container.database\" is not defined"):
        container.user_service()


def test_call_undefined_error_message_with_container_provider_parent_deep():
    class Database:
        pass

    class UserService:
        def __init__(self, db):
            self.db = db

    class Gateways(containers.DeclarativeContainer):
        database_client = providers.Singleton(Database)

    class Services(containers.DeclarativeContainer):
        gateways = providers.DependenciesContainer()

        user = providers.Factory(
            UserService,
            db=gateways.database_client,
        )

    class Container(containers.DeclarativeContainer):
        gateways = providers.Container(Gateways)

        services = providers.Container(
            Services,
            # gateways=gateways,  # <---- missing dependency
        )

    container = Container()

    with raises(errors.Error, match="Dependency \"Container.services.gateways.database_client\" is not defined"):
        container.services().user()


def test_call_undefined_error_message_with_dependenciescontainer_provider_parent():
    class UserService:
        def __init__(self, db):
            self.db = db

    class Services(containers.DeclarativeContainer):
        gateways = providers.DependenciesContainer()

        user = providers.Factory(
            UserService,
            db=gateways.database_client,  # <---- missing dependency
        )

    services = Services()

    with raises(errors.Error, match="Dependency \"Services.gateways.database_client\" is not defined"):
        services.user()


def test_assign_parent(provider):
    parent = providers.DependenciesContainer()
    provider.assign_parent(parent)
    assert provider.parent is parent


def test_parent_name(provider):
    container = containers.DynamicContainer()
    container.name = provider
    assert provider.parent_name == "name"


def test_parent_name_with_deep_parenting(provider):
    container = providers.DependenciesContainer(name=provider)
    _ = providers.DependenciesContainer(container=container)
    assert provider.parent_name == "container.name"


def test_parent_name_is_none():
    provider = providers.Dependency()
    assert provider.parent_name is None


def test_parent_deepcopy(provider):
    container = containers.DynamicContainer()
    container.name = provider

    copied = providers.deepcopy(container)

    assert container.name.parent is container
    assert copied.name.parent is copied

    assert container is not copied
    assert container.name is not copied.name
    assert container.name.parent is not copied.name.parent


def test_forward_attr_to_default():
    default = providers.Configuration()
    provider = providers.Dependency(default=default)
    provider.from_dict({"foo": "bar"})
    assert default() == {"foo": "bar"}


def test_forward_attr_to_overriding(provider):
    overriding = providers.Configuration()
    provider.override(overriding)
    provider.from_dict({"foo": "bar"})
    assert overriding() == {"foo": "bar"}


def test_forward_attr_to_none(provider):
    with raises(AttributeError):
        provider.from_dict


def test_deepcopy(provider):
    provider_copy = providers.deepcopy(provider)
    assert provider is not provider_copy
    assert isinstance(provider, providers.Dependency)


def test_deepcopy_from_memo(provider):
    provider_copy_memo = providers.Provider()
    provider_copy = providers.deepcopy(provider, memo={id(provider): provider_copy_memo})
    assert provider_copy is provider_copy_memo


def test_deepcopy_overridden(provider):
    overriding_provider = providers.Provider()

    provider.override(overriding_provider)

    provider_copy = providers.deepcopy(provider)
    overriding_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert isinstance(provider, providers.Dependency)

    assert overriding_provider is not overriding_provider_copy
    assert isinstance(overriding_provider_copy, providers.Provider)


def test_deep_copy_default_object():
    default = {"foo": "bar"}
    provider = providers.Dependency(dict, default=default)

    provider_copy = providers.deepcopy(provider)

    assert provider_copy() is default
    assert provider_copy.default() is default


def test_deep_copy_default_provider():
    bar = object()
    default = providers.Factory(dict, foo=providers.Object(bar))
    provider = providers.Dependency(dict, default=default)

    provider_copy = providers.deepcopy(provider)

    assert provider_copy() == {"foo": bar}
    assert provider_copy.default() == {"foo": bar}
    assert provider_copy()["foo"] is bar


def test_with_container_default_object():
    default = {"foo": "bar"}

    class Container(containers.DeclarativeContainer):
        provider = providers.Dependency(dict, default=default)

    container = Container()

    assert container.provider() is default
    assert container.provider.default() is default


def test_with_container_default_provider():
    bar = object()

    class Container(containers.DeclarativeContainer):
        provider = providers.Dependency(dict, default=providers.Factory(dict, foo=providers.Object(bar)))

    container = Container()

    assert container.provider() == {"foo": bar}
    assert container.provider.default() == {"foo": bar}
    assert container.provider()["foo"] is bar


def test_with_container_default_provider_with_overriding():
    bar = object()
    baz = object()

    class Container(containers.DeclarativeContainer):
        provider = providers.Dependency(dict, default=providers.Factory(dict, foo=providers.Object(bar)))

    container = Container(provider=providers.Factory(dict, foo=providers.Object(baz)))

    assert container.provider() == {"foo": baz}
    assert container.provider.default() == {"foo": bar}
    assert container.provider()["foo"] is baz


def test_repr(provider):
    assert repr(provider) == (
        "<dependency_injector.providers."
        "Dependency({0}) at {1}>".format(repr(list), hex(id(provider)))
    )


def test_repr_in_container():
    class Container(containers.DeclarativeContainer):
        dependency = providers.Dependency(instance_of=int)

    container = Container()

    assert repr(container.dependency) == (
        "<dependency_injector.providers."
        "Dependency({0}) at {1}, container name: \"Container.dependency\">".format(
            repr(int),
            hex(id(container.dependency)),
        )
    )


def test_external_dependency():
    assert isinstance(providers.ExternalDependency(), providers.Dependency)



import asyncio

from dependency_injector import containers, providers
from dependency_injector.wiring import inject, Provide, Closing


class TestResource:
    def __init__(self):
        self.init_counter = 0
        self.shutdown_counter = 0

    def reset_counters(self):
        self.init_counter = 0
        self.shutdown_counter = 0


resource1 = TestResource()
resource2 = TestResource()


async def async_resource(resource):
    await asyncio.sleep(0.001)
    resource.init_counter += 1

    yield resource

    await asyncio.sleep(0.001)
    resource.shutdown_counter += 1


class Container(containers.DeclarativeContainer):

    resource1 = providers.Resource(async_resource, providers.Object(resource1))
    resource2 = providers.Resource(async_resource, providers.Object(resource2))


@inject
async def async_injection(
        resource1: object = Provide[Container.resource1],
        resource2: object = Provide[Container.resource2],
):
    return resource1, resource2


@inject
async def async_injection_with_closing(
        resource1: object = Closing[Provide[Container.resource1]],
        resource2: object = Closing[Provide[Container.resource2]],
):
    return resource1, resource2



"""ProvidedInstance provider tests."""

from dependency_injector import containers, providers
from pytest import fixture


class Service:
    def __init__(self, value):
        self.value = value
        self.values = [self.value]

    def __call__(self):
        return self.value

    def __getitem__(self, item):
        return self.values[item]

    def get_value(self):
        return self.value

    def get_closure(self):
        def closure():
            return self.value
        return closure


class Client:
    def __init__(self, value):
        self.value = value


class Container(containers.DeclarativeContainer):

    service = providers.Singleton(Service, value="foo")

    client_attribute = providers.Factory(
        Client,
        value=service.provided.value,
    )

    client_item = providers.Factory(
        Client,
        value=service.provided[0],
    )

    client_attribute_item = providers.Factory(
        Client,
        value=service.provided.values[0],
    )

    client_method_call = providers.Factory(
        Client,
        value=service.provided.get_value.call(),
    )
    client_method_closure_call = providers.Factory(
        Client,
        value=service.provided.get_closure.call().call(),
    )

    client_provided_call = providers.Factory(
        Client,
        value=service.provided.call(),
    )


@fixture
def container():
    return Container()


def test_is_provider(container):
    assert providers.is_provider(container.service.provided) is True


def test_attribute(container):
    client = container.client_attribute()
    assert client.value == "foo"


def test_item(container):
    client = container.client_item()
    assert client.value == "foo"


def test_attribute_item(container):
    client = container.client_attribute_item()
    assert client.value == "foo"


def test_method_call(container):
    client = container.client_method_call()
    assert client.value == "foo"


def test_method_closure_call(container):
    client = container.client_method_closure_call()
    assert client.value == "foo"


def test_provided_call(container):
    client = container.client_provided_call()
    assert client.value == "foo"


def test_call_overridden(container):
    value = "bar"
    with container.service.override(Service(value)):
        assert container.client_attribute().value == value
        assert container.client_item().value == value
        assert container.client_attribute_item().value == value
        assert container.client_method_call().value == value


def test_repr_provided_instance(container):
    provider = container.service.provided
    assert repr(provider) == "ProvidedInstance(\"{0}\")".format(repr(container.service))


def test_repr_attribute_getter(container):
    provider = container.service.provided.value
    assert repr(provider) == "AttributeGetter(\"value\")"


def test_repr_item_getter(container):
    provider = container.service.provided["test-test"]
    assert repr(provider) == "ItemGetter(\"test-test\")"


def test_provided_instance():
    provides = providers.Object(object())
    provider = providers.ProvidedInstance()
    provider.set_provides(provides)
    assert provider.provides is provides
    assert provider.set_provides(providers.Provider()) is provider


def test_attribute_getter():
    provides = providers.Object(object())
    provider = providers.AttributeGetter()
    provider.set_provides(provides)
    provider.set_name("__dict__")
    assert provider.provides is provides
    assert provider.name == "__dict__"
    assert provider.set_provides(providers.Provider()) is provider
    assert provider.set_name("__dict__") is provider


def test_item_getter():
    provides = providers.Object({"foo": "bar"})
    provider = providers.ItemGetter()
    provider.set_provides(provides)
    provider.set_name("foo")
    assert provider.provides is provides
    assert provider.name == "foo"
    assert provider.set_provides(providers.Provider()) is provider
    assert provider.set_name("foo") is provider


def test_method_caller():
    provides = providers.Object(lambda: 42)
    provider = providers.MethodCaller()
    provider.set_provides(provides)
    assert provider.provides is provides
    assert provider() == 42
    assert provider.set_provides(providers.Provider()) is provider


def test_puzzled():
    service = providers.Singleton(Service, value="foo-bar")

    dependency = providers.Object(
        {
            "a": {
                "b": {
                    "c1": 10,
                    "c2": lambda arg: {"arg": arg}
                },
            },
        },
    )

    test_list = providers.List(
        dependency.provided["a"]["b"]["c1"],
        dependency.provided["a"]["b"]["c2"].call(22)["arg"],
        dependency.provided["a"]["b"]["c2"].call(service)["arg"],
        dependency.provided["a"]["b"]["c2"].call(service)["arg"].value,
        dependency.provided["a"]["b"]["c2"].call(service)["arg"].get_value.call(),
    )

    result = test_list()
    assert result == [
        10,
        22,
        service(),
        "foo-bar",
        "foo-bar",
    ]


def test_provided_attribute_in_base_class():
    provider = providers.Provider()
    assert isinstance(provider.provided, providers.ProvidedInstance)



"""Provider utils tests."""

from dependency_injector import providers, errors
from pytest import raises


def test_with_instance():
    provider = providers.Provider()
    assert providers.ensure_is_provider(provider), provider


def test_with_class():
    with raises(errors.Error):
        providers.ensure_is_provider(providers.Provider)


def test_with_string():
    with raises(errors.Error):
        providers.ensure_is_provider("some_string")


def test_with_object():
    with raises(errors.Error):
        providers.ensure_is_provider(object())



"""Flask extension module."""

from __future__ import absolute_import
import warnings

from flask import request as flask_request

from dependency_injector import providers, errors


warnings.warn(
    'Module "dependency_injector.ext.flask" is deprecated since '
    'version 4.0.0. Use "dependency_injector.wiring" module instead.',
    category=DeprecationWarning,
)


request = providers.Object(flask_request)


class Application(providers.Singleton):
    """Flask application provider."""


class Extension(providers.Singleton):
    """Flask extension provider."""


class View(providers.Callable):
    """Flask view provider."""

    def as_view(self):
        """Return Flask view function."""
        return as_view(self)


class ClassBasedView(providers.Factory):
    """Flask class-based view provider."""

    def as_view(self, name):
        """Return Flask view function."""
        return as_view(self, name)


def as_view(provider, name=None):
    """Transform class-based view provider to view function."""
    if isinstance(provider, providers.Factory):
        def view(*args, **kwargs):
            self = provider()
            return self.dispatch_request(*args, **kwargs)

        assert name, 'Argument "endpoint" is required for class-based views'
        view.__name__ = name
    elif isinstance(provider, providers.Callable):
        def view(*args, **kwargs):
            return provider(*args, **kwargs)

        view.__name__ = provider.provides.__name__
    else:
        raise errors.Error('Undefined provider type')

    view.__doc__ = provider.provides.__doc__
    view.__module__ = provider.provides.__module__

    if isinstance(provider.provides, type):
        view.view_class = provider.provides

    if hasattr(provider.provides, 'decorators'):
        for decorator in provider.provides.decorators:
            view = decorator(view)

    if hasattr(provider.provides, 'methods'):
        view.methods = provider.provides.methods

    if hasattr(provider.provides, 'provide_automatic_options'):
        view.provide_automatic_options = provider.provides.provide_automatic_options

    return view



"""Test that wiring works properly with @functools.wraps decorator.

See issue for details: https://github.com/ets-labs/python-dependency-injector/issues/454
"""

import functools

from dependency_injector.wiring import inject, Provide
from pytest import fixture

from samples.wiring.container import Container


@fixture
def container():
    container = Container()
    yield container
    container.unwire()


def decorator1(func):
    @functools.wraps(func)
    @inject
    def wrapper(value1: int = Provide[Container.config.value1]):
        result = func()
        return result + value1
    return wrapper


def decorator2(func):
    @functools.wraps(func)
    @inject
    def wrapper(value2: int = Provide[Container.config.value2]):
        result = func()
        return result + value2
    return wrapper


@decorator1
@decorator2
def sample():
    return 2


def test_wraps(container: Container):
    container.wire(modules=[__name__])
    container.config.from_dict({"value1": 42, "value2": 15})

    assert sample() == 2 + 42 + 15



"""Common test artifacts."""


class Example:
    def __init__(self, init_arg1=None, init_arg2=None, init_arg3=None, init_arg4=None):
        self.init_arg1 = init_arg1
        self.init_arg2 = init_arg2
        self.init_arg3 = init_arg3
        self.init_arg4 = init_arg4

        self.attribute1 = None
        self.attribute2 = None


class ExampleA(Example):
    pass


class ExampleB(Example):
    pass



import sys

if sys.version_info >= (3, 6):
    from dependency_injector.wiring import Provide

    from ..container import Container
    from ..service import Service


    def test_package_function(service: Service = Provide[Container.service]):
        return service



"""AbstractSingleton provider tests."""

from dependency_injector import providers, errors
from pytest import raises

from .common import Example


def test_inheritance():
    assert isinstance(providers.AbstractSingleton(Example), providers.BaseSingleton)


def test_call_overridden_by_singleton():
    provider = providers.AbstractSingleton(object)
    provider.override(providers.Singleton(Example))
    assert isinstance(provider(), Example)


def test_call_overridden_by_delegated_singleton():
    provider = providers.AbstractSingleton(object)
    provider.override(providers.DelegatedSingleton(Example))
    assert isinstance(provider(), Example)


def test_call_not_overridden():
    provider = providers.AbstractSingleton(object)
    with raises(errors.Error):
        provider()


def test_reset_overridden():
    provider = providers.AbstractSingleton(object)
    provider.override(providers.Singleton(Example))

    instance1 = provider()

    provider.reset()

    instance2 = provider()

    assert instance1 is not instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_reset_not_overridden():
    provider = providers.AbstractSingleton(object)
    with raises(errors.Error):
        provider.reset()


def test_override_by_not_singleton():
    provider = providers.AbstractSingleton(object)
    with raises(errors.Error):
        provider.override(providers.Factory(object))


def test_repr():
    provider = providers.AbstractSingleton(Example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "AbstractSingleton({0}) at {1}>".format(repr(Example), hex(id(provider)))
    )



"""Configuration.from_pydantic() tests."""

from pydantic import BaseModel

try:
    from pydantic_settings import (
        BaseSettings,  # type: ignore[import-not-found,unused-ignore]
    )
except ImportError:
    try:
        from pydantic import BaseSettings  # type: ignore[no-redef,unused-ignore]
    except ImportError:

        class BaseSettings:  # type: ignore[no-redef]
            """No-op fallback"""


from pytest import fixture, mark, raises

from dependency_injector import providers

pytestmark = mark.pydantic


class Section11(BaseModel):
    value1: int = 1


class Section12(BaseModel):
    value2: int = 2


class Settings1(BaseSettings):
    section1: Section11 = Section11()
    section2: Section12 = Section12()


class Section21(BaseModel):
    value1: int = 11
    value11: int = 11


class Section3(BaseModel):
    value3: int = 3


class Settings2(BaseSettings):
    section1: Section21 = Section21()
    section3: Section3 = Section3()


@fixture
def config(config_type, pydantic_settings_1, pydantic_settings_2):
    if config_type == "strict":
        return providers.Configuration(strict=True)
    elif config_type == "default":
        return providers.Configuration(pydantic_settings=[pydantic_settings_1, pydantic_settings_2])
    else:
        raise ValueError("Undefined config type \"{0}\"".format(config_type))


@fixture
def pydantic_settings_1():
    return Settings1()


@fixture
def pydantic_settings_2():
    return Settings2()


def test_load(config):
    config.load()

    assert config() == {
        "section1": {
            "value1": 11,
            "value11": 11,
        },
        "section2": {
            "value2": 2,
        },
        "section3": {
            "value3": 3,
        },
    }
    assert config.section1() == {"value1": 11, "value11": 11}
    assert config.section1.value1() == 11
    assert config.section1.value11() == 11
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2
    assert config.section3() == {"value3": 3}
    assert config.section3.value3() == 3


def test_get_pydantic_settings(config, pydantic_settings_1, pydantic_settings_2):
    assert config.get_pydantic_settings() == [pydantic_settings_1, pydantic_settings_2]


def test_copy(config, pydantic_settings_1, pydantic_settings_2):
    config_copy = providers.deepcopy(config)
    assert config_copy.get_pydantic_settings() == [pydantic_settings_1, pydantic_settings_2]


def test_set_pydantic_settings(config):
    class Settings3(BaseSettings):
        ...

    class Settings4(BaseSettings):
        ...

    settings_3 = Settings3()
    settings_4 = Settings4()

    config.set_pydantic_settings([settings_3, settings_4])
    assert config.get_pydantic_settings() == [settings_3, settings_4]


def test_file_does_not_exist(config):
    config.set_pydantic_settings([BaseSettings()])
    config.load()
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_file_does_not_exist_strict_mode(config):
    config.set_pydantic_settings([BaseSettings()])
    with raises(ValueError):
        config.load()
    assert config() == {}


def test_required_file_does_not_exist(config):
    config.set_pydantic_settings([BaseSettings()])
    with raises(ValueError):
        config.load(required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_file_does_not_exist_strict_mode(config):
    config.set_pydantic_settings([BaseSettings()])
    config.load(required=False)
    assert config() == {}



"""Tests for provide async mode typing stubs."""

from pytest import mark

from .common import Container, Client, Service, RESOURCE1, RESOURCE2


@mark.asyncio
async def test_async_():
    container = Container()

    client1 = await container.client.async_()
    client2 = await container.client.async_()

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is RESOURCE2

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is RESOURCE2

    service1 = await container.service.async_()
    service2 = await container.service.async_()

    assert isinstance(service1, Service)
    assert isinstance(service1.client, Client)
    assert service1.client.resource1 is RESOURCE1
    assert service1.client.resource2 is RESOURCE2

    assert isinstance(service2, Service)
    assert isinstance(service2.client, Client)
    assert service2.client.resource1 is RESOURCE1
    assert service2.client.resource2 is RESOURCE2

    assert service1.client is not service2.client



"""List provider async mode tests."""

from dependency_injector import containers, providers
from pytest import mark


@mark.asyncio
async def test_provide():
    # See issue: https://github.com/ets-labs/python-dependency-injector/issues/450
    async def create_resource(param: str):
        return param

    class Container(containers.DeclarativeContainer):

        resources = providers.List(
            providers.Resource(create_resource, "foo"),
            providers.Resource(create_resource, "bar")
        )

    container = Container()
    resources = await container.resources()

    assert resources[0] == "foo"
    assert resources[1] == "bar"



"""Test module for wiring."""

import sys

if "pypy" not in sys.version.lower():
    import numpy  # noqa
    from numpy import *  # noqa
    if sys.version_info >= (3, 7):
        from numpy.typing import *  # noqa

    import scipy  # noqa
    from scipy import *  # noqa

    import builtins  # noqa
    from builtins import *  # noqa



from typing import Any

from dependency_injector import providers


# Test 1: to check the return type
provider1 = providers.Selector(
    lambda: "a",
    a=providers.Factory(object),
    b=providers.Factory(object),
)
var1: Any = provider1()

# Test 2: to check the provided instance interface
provider2 = providers.Selector(
    lambda: "a",
    a=providers.Factory(object),
    b=providers.Factory(object),
)
provided2: providers.ProvidedInstance = provider2.provided
attr_getter2: providers.AttributeGetter = provider2.provided.attr
item_getter2: providers.ItemGetter = provider2.provided["item"]
method_caller2: providers.MethodCaller = provider2.provided.method.call(123, arg=324)

# Test3 to check the getattr
provider3 = providers.Selector(
    lambda: "a",
    a=providers.Factory(object),
    b=providers.Factory(object),
)
attr3: providers.Provider = provider3.a

# Test 4: to check the return type with await
provider4 = providers.Selector(
    lambda: "a",
    a=providers.Factory(object),
    b=providers.Factory(object),
)
async def _async4() -> None:
    var1: Any = await provider4()  # type: ignore
    var2: Any = await provider4.async_()



"""Tests for provider async mode API."""

from dependency_injector import providers
from pytest import fixture


@fixture
def provider():
    return providers.Provider()


def test_default_mode(provider: providers.Provider):
    assert provider.is_async_mode_enabled() is False
    assert provider.is_async_mode_disabled() is False
    assert provider.is_async_mode_undefined() is True


def test_enable(provider: providers.Provider):
    provider.enable_async_mode()

    assert provider.is_async_mode_enabled() is True
    assert provider.is_async_mode_disabled() is False
    assert provider.is_async_mode_undefined() is False


def test_disable(provider: providers.Provider):
    provider.disable_async_mode()

    assert provider.is_async_mode_enabled() is False
    assert provider.is_async_mode_disabled() is True
    assert provider.is_async_mode_undefined() is False


def test_reset(provider: providers.Provider):
    provider.enable_async_mode()

    assert provider.is_async_mode_enabled() is True
    assert provider.is_async_mode_disabled() is False
    assert provider.is_async_mode_undefined() is False

    provider.reset_async_mode()

    assert provider.is_async_mode_enabled() is False
    assert provider.is_async_mode_disabled() is False
    assert provider.is_async_mode_undefined() is True



from dependency_injector import containers, providers

from .service import Service


class SubContainer(containers.DeclarativeContainer):

    int_object = providers.Object(1)


class Container(containers.DeclarativeContainer):

    config = providers.Configuration()

    service = providers.Factory(Service)

    sub = providers.Container(SubContainer)



"""Wiring module."""

import functools
import inspect
import importlib
import importlib.machinery
import pkgutil
import warnings
import sys
from types import ModuleType
from typing import (
    Optional,
    Iterable,
    Iterator,
    Callable,
    Any,
    Tuple,
    Dict,
    Generic,
    TypeVar,
    Type,
    Union,
    Set,
    cast,
)

if sys.version_info < (3, 7):
    from typing import GenericMeta
else:
    class GenericMeta(type):
        ...

# Hotfix, see: https://github.com/ets-labs/python-dependency-injector/issues/362
if sys.version_info >= (3, 9):
    from types import GenericAlias
else:
    GenericAlias = None


try:
    import fastapi.params
except ImportError:
    fastapi = None


try:
    import starlette.requests
except ImportError:
    starlette = None


try:
    import werkzeug.local
except ImportError:
    werkzeug = None


from . import providers

if sys.version_info[:2] == (3, 5):
    warnings.warn(
        "Dependency Injector will drop support of Python 3.5 after Jan 1st of 2022. "
        "This does not mean that there will be any immediate breaking changes, "
        "but tests will no longer be executed on Python 3.5, and bugs will not be addressed.",
        category=DeprecationWarning,
    )

__all__ = (
    "wire",
    "unwire",
    "inject",
    "as_int",
    "as_float",
    "as_",
    "required",
    "invariant",
    "provided",
    "Provide",
    "Provider",
    "Closing",
    "register_loader_containers",
    "unregister_loader_containers",
    "install_loader",
    "uninstall_loader",
    "is_loader_installed",
)

T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])
Container = Any


class PatchedRegistry:

    def __init__(self) -> None:
        self._callables: Dict[Callable[..., Any], "PatchedCallable"] = {}
        self._attributes: Set[PatchedAttribute] = set()

    def register_callable(self, patched: "PatchedCallable") -> None:
        self._callables[patched.patched] = patched

    def get_callables_from_module(self, module: ModuleType) -> Iterator[Callable[..., Any]]:
        for patched_callable in self._callables.values():
            if not patched_callable.is_in_module(module):
                continue
            yield patched_callable.patched

    def get_callable(self, fn: Callable[..., Any]) -> "PatchedCallable":
        return self._callables.get(fn)

    def has_callable(self, fn: Callable[..., Any]) -> bool:
        return fn in self._callables

    def register_attribute(self, patched: "PatchedAttribute") -> None:
        self._attributes.add(patched)

    def get_attributes_from_module(self, module: ModuleType) -> Iterator["PatchedAttribute"]:
        for attribute in self._attributes:
            if not attribute.is_in_module(module):
                continue
            yield attribute

    def clear_module_attributes(self, module: ModuleType) -> None:
        for attribute in self._attributes.copy():
            if not attribute.is_in_module(module):
                continue
            self._attributes.remove(attribute)


class PatchedCallable:

    __slots__ = (
        "patched",
        "original",
        "reference_injections",
        "injections",
        "reference_closing",
        "closing",
    )

    def __init__(
            self,
            patched: Optional[Callable[..., Any]] = None,
            original: Optional[Callable[..., Any]] = None,
            reference_injections: Optional[Dict[Any, Any]] = None,
            reference_closing: Optional[Dict[Any, Any]] = None,
    ) -> None:
        self.patched = patched
        self.original = original

        if reference_injections is None:
            reference_injections = {}
        self.reference_injections: Dict[Any, Any] = reference_injections.copy()
        self.injections: Dict[Any, Any] = {}

        if reference_closing is None:
            reference_closing = {}
        self.reference_closing: Dict[Any, Any] = reference_closing.copy()
        self.closing: Dict[Any, Any] = {}

    def is_in_module(self, module: ModuleType) -> bool:
        if self.patched is None:
            return False
        return self.patched.__module__ == module.__name__

    def add_injection(self, kwarg: Any, injection: Any) -> None:
        self.injections[kwarg] = injection

    def add_closing(self, kwarg: Any, injection: Any) -> None:
        self.closing[kwarg] = injection

    def unwind_injections(self) -> None:
        self.injections = {}
        self.closing = {}


class PatchedAttribute:

    __slots__ = (
        "member",
        "name",
        "marker",
    )

    def __init__(self, member: Any, name: str, marker: "_Marker") -> None:
        self.member = member
        self.name = name
        self.marker = marker

    @property
    def module_name(self) -> str:
        if isinstance(self.member, ModuleType):
            return self.member.__name__
        else:
            return self.member.__module__

    def is_in_module(self, module: ModuleType) -> bool:
        return self.module_name == module.__name__


class ProvidersMap:

    CONTAINER_STRING_ID = "<container>"

    def __init__(self, container) -> None:
        self._container = container
        self._map = self._create_providers_map(
            current_container=container,
            original_container=(
                container.declarative_parent
                if container.declarative_parent
                else container
            ),
        )

    def resolve_provider(
            self,
            provider: Union[providers.Provider, str],
            modifier: Optional["Modifier"] = None,
    ) -> Optional[providers.Provider]:
        if isinstance(provider, providers.Delegate):
            return self._resolve_delegate(provider)
        elif isinstance(provider, (
            providers.ProvidedInstance,
            providers.AttributeGetter,
            providers.ItemGetter,
            providers.MethodCaller,
        )):
            return self._resolve_provided_instance(provider)
        elif isinstance(provider, providers.ConfigurationOption):
            return self._resolve_config_option(provider)
        elif isinstance(provider, providers.TypedConfigurationOption):
            return self._resolve_config_option(provider.option, as_=provider.provides)
        elif isinstance(provider, str):
            return self._resolve_string_id(provider, modifier)
        else:
            return self._resolve_provider(provider)

    def _resolve_string_id(
            self,
            id: str,
            modifier: Optional["Modifier"] = None,
    ) -> Optional[providers.Provider]:
        if id == self.CONTAINER_STRING_ID:
            return self._container.__self__

        provider = self._container
        for segment in id.split("."):
            try:
                provider = getattr(provider, segment)
            except AttributeError:
                return None

        if modifier:
            provider = modifier.modify(provider, providers_map=self)
        return provider

    def _resolve_provided_instance(
            self,
            original: providers.Provider,
    ) -> Optional[providers.Provider]:
        modifiers = []
        while isinstance(original, (
                providers.ProvidedInstance,
                providers.AttributeGetter,
                providers.ItemGetter,
                providers.MethodCaller,
        )):
            modifiers.insert(0, original)
            original = original.provides

        new = self._resolve_provider(original)
        if new is None:
            return None

        for modifier in modifiers:
            if isinstance(modifier, providers.ProvidedInstance):
                new = new.provided
            elif isinstance(modifier, providers.AttributeGetter):
                new = getattr(new, modifier.name)
            elif isinstance(modifier, providers.ItemGetter):
                new = new[modifier.name]
            elif isinstance(modifier, providers.MethodCaller):
                new = new.call(
                    *modifier.args,
                    **modifier.kwargs,
                )

        return new

    def _resolve_delegate(
            self,
            original: providers.Delegate,
    ) -> Optional[providers.Provider]:
        provider = self._resolve_provider(original.provides)
        if provider:
            provider = provider.provider
        return provider

    def _resolve_config_option(
            self,
            original: providers.ConfigurationOption,
            as_: Any = None,
    ) -> Optional[providers.Provider]:
        original_root = original.root
        new = self._resolve_provider(original_root)
        if new is None:
            return None
        new = cast(providers.Configuration, new)

        for segment in original.get_name_segments():
            if providers.is_provider(segment):
                segment = self.resolve_provider(segment)
                new = new[segment]
            else:
                new = getattr(new, segment)

        if original.is_required():
            new = new.required()

        if as_:
            new = new.as_(as_)

        return new

    def _resolve_provider(
            self,
            original: providers.Provider,
    ) -> Optional[providers.Provider]:
        try:
            return self._map[original]
        except KeyError:
            return None

    @classmethod
    def _create_providers_map(
            cls,
            current_container: Container,
            original_container: Container,
    ) -> Dict[providers.Provider, providers.Provider]:
        current_providers = current_container.providers
        current_providers["__self__"] = current_container.__self__

        original_providers = original_container.providers
        original_providers["__self__"] = original_container.__self__

        providers_map = {}
        for provider_name, current_provider in current_providers.items():
            original_provider = original_providers[provider_name]
            providers_map[original_provider] = current_provider

            if isinstance(current_provider, providers.Container) \
                    and isinstance(original_provider, providers.Container):
                subcontainer_map = cls._create_providers_map(
                    current_container=current_provider.container,
                    original_container=original_provider.container,
                )
                providers_map.update(subcontainer_map)

        return providers_map


class InspectFilter:

    def is_excluded(self, instance: object) -> bool:
        if self._is_werkzeug_local_proxy(instance):
            return True
        elif self._is_starlette_request_cls(instance):
            return True
        elif self._is_builtin(instance):
            return True
        else:
            return False

    def _is_werkzeug_local_proxy(self, instance: object) -> bool:
        return werkzeug and isinstance(instance, werkzeug.local.LocalProxy)

    def _is_starlette_request_cls(self, instance: object) -> bool:
        return starlette \
               and isinstance(instance, type) \
               and _safe_is_subclass(instance, starlette.requests.Request)

    def _is_builtin(self, instance: object) -> bool:
        return inspect.isbuiltin(instance)


def wire(  # noqa: C901
        container: Container,
        *,
        modules: Optional[Iterable[ModuleType]] = None,
        packages: Optional[Iterable[ModuleType]] = None,
) -> None:
    """Wire container providers with provided packages and modules."""
    modules = [*modules] if modules else []

    if packages:
        for package in packages:
            modules.extend(_fetch_modules(package))

    providers_map = ProvidersMap(container)

    for module in modules:
        for member_name, member in inspect.getmembers(module):
            if _inspect_filter.is_excluded(member):
                continue

            if _is_marker(member):
                _patch_attribute(module, member_name, member, providers_map)
            elif inspect.isfunction(member):
                _patch_fn(module, member_name, member, providers_map)
            elif inspect.isclass(member):
                cls = member
                try:
                    cls_members = inspect.getmembers(cls)
                except Exception:  # noqa
                    # Hotfix, see: https://github.com/ets-labs/python-dependency-injector/issues/441
                    continue
                else:
                    for cls_member_name, cls_member in cls_members:
                        if _is_marker(cls_member):
                            _patch_attribute(cls, cls_member_name, cls_member, providers_map)
                        elif _is_method(cls_member):
                            _patch_method(cls, cls_member_name, cls_member, providers_map)

        for patched in _patched_registry.get_callables_from_module(module):
            _bind_injections(patched, providers_map)


def unwire(  # noqa: C901
        *,
        modules: Optional[Iterable[ModuleType]] = None,
        packages: Optional[Iterable[ModuleType]] = None,
) -> None:
    """Wire provided packages and modules with previous wired providers."""
    modules = [*modules] if modules else []

    if packages:
        for package in packages:
            modules.extend(_fetch_modules(package))

    for module in modules:
        for name, member in inspect.getmembers(module):
            if inspect.isfunction(member):
                _unpatch(module, name, member)
            elif inspect.isclass(member):
                for method_name, method in inspect.getmembers(member, inspect.isfunction):
                    _unpatch(member, method_name, method)

        for patched in _patched_registry.get_callables_from_module(module):
            _unbind_injections(patched)

        for patched_attribute in _patched_registry.get_attributes_from_module(module):
            _unpatch_attribute(patched_attribute)
        _patched_registry.clear_module_attributes(module)


def inject(fn: F) -> F:
    """Decorate callable with injecting decorator."""
    reference_injections, reference_closing = _fetch_reference_injections(fn)
    patched = _get_patched(fn, reference_injections, reference_closing)
    return cast(F, patched)


def _patch_fn(
        module: ModuleType,
        name: str,
        fn: Callable[..., Any],
        providers_map: ProvidersMap,
) -> None:
    if not _is_patched(fn):
        reference_injections, reference_closing = _fetch_reference_injections(fn)
        if not reference_injections:
            return
        fn = _get_patched(fn, reference_injections, reference_closing)

    _bind_injections(fn, providers_map)

    setattr(module, name, fn)


def _patch_method(
        cls: Type,
        name: str,
        method: Callable[..., Any],
        providers_map: ProvidersMap,
) -> None:
    if hasattr(cls, "__dict__") \
            and name in cls.__dict__ \
            and isinstance(cls.__dict__[name], (classmethod, staticmethod)):
        method = cls.__dict__[name]
        fn = method.__func__
    else:
        fn = method

    if not _is_patched(fn):
        reference_injections, reference_closing = _fetch_reference_injections(fn)
        if not reference_injections:
            return
        fn = _get_patched(fn, reference_injections, reference_closing)

    _bind_injections(fn, providers_map)

    if isinstance(method, (classmethod, staticmethod)):
        fn = type(method)(fn)

    setattr(cls, name, fn)


def _unpatch(
        module: ModuleType,
        name: str,
        fn: Callable[..., Any],
) -> None:
    if hasattr(module, "__dict__") \
            and name in module.__dict__ \
            and isinstance(module.__dict__[name], (classmethod, staticmethod)):
        method = module.__dict__[name]
        fn = method.__func__

    if not _is_patched(fn):
        return

    _unbind_injections(fn)


def _patch_attribute(
        member: Any,
        name: str,
        marker: "_Marker",
        providers_map: ProvidersMap,
) -> None:
    provider = providers_map.resolve_provider(marker.provider, marker.modifier)
    if provider is None:
        return

    _patched_registry.register_attribute(PatchedAttribute(member, name, marker))

    if isinstance(marker, Provide):
        instance = provider()
        setattr(member, name, instance)
    elif isinstance(marker, Provider):
        setattr(member, name, provider)
    else:
        raise Exception(f"Unknown type of marker {marker}")


def _unpatch_attribute(patched: PatchedAttribute) -> None:
    setattr(patched.member, patched.name, patched.marker)


def _fetch_reference_injections(  # noqa: C901
        fn: Callable[..., Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    # Hotfix, see:
    # - https://github.com/ets-labs/python-dependency-injector/issues/362
    # - https://github.com/ets-labs/python-dependency-injector/issues/398
    if GenericAlias and any((
                fn is GenericAlias,
                getattr(fn, "__func__", None) is GenericAlias
            )):
        fn = fn.__init__

    try:
        signature = inspect.signature(fn)
    except ValueError as exception:
        if "no signature found" in str(exception):
            return {}, {}
        elif "not supported by signature" in str(exception):
            return {}, {}
        else:
            raise exception

    injections = {}
    closing = {}
    for parameter_name, parameter in signature.parameters.items():
        if not isinstance(parameter.default, _Marker) \
                and not _is_fastapi_depends(parameter.default):
            continue

        marker = parameter.default

        if _is_fastapi_depends(marker):
            marker = marker.dependency

            if not isinstance(marker, _Marker):
                continue

        if isinstance(marker, Closing):
            marker = marker.provider
            closing[parameter_name] = marker

        injections[parameter_name] = marker
    return injections, closing


def _locate_dependent_closing_args(provider: providers.Provider) -> Dict[str, providers.Provider]:
    if not hasattr(provider, "args"):
        return {}

    closing_deps = {}
    for arg in provider.args:
        if not isinstance(arg, providers.Provider) or not hasattr(arg, "args"):
            continue

        if not arg.args and isinstance(arg, providers.Resource):
            return {str(id(arg)): arg}
        else:
            closing_deps += _locate_dependent_closing_args(arg)
    return closing_deps


def _bind_injections(fn: Callable[..., Any], providers_map: ProvidersMap) -> None:
    patched_callable = _patched_registry.get_callable(fn)
    if patched_callable is None:
        return

    for injection, marker in patched_callable.reference_injections.items():
        provider = providers_map.resolve_provider(marker.provider, marker.modifier)

        if provider is None:
            continue

        if isinstance(marker, Provide):
            patched_callable.add_injection(injection, provider)
        elif isinstance(marker, Provider):
            if isinstance(provider, providers.Delegate):
                patched_callable.add_injection(injection, provider)
            else:
                patched_callable.add_injection(injection, provider.provider)

        if injection in patched_callable.reference_closing:
            patched_callable.add_closing(injection, provider)
            deps = _locate_dependent_closing_args(provider)
            for key, dep in deps.items():
                patched_callable.add_closing(key, dep)


def _unbind_injections(fn: Callable[..., Any]) -> None:
    patched_callable = _patched_registry.get_callable(fn)
    if patched_callable is None:
        return
    patched_callable.unwind_injections()


def _fetch_modules(package):
    modules = [package]
    if not hasattr(package, "__path__") or not hasattr(package, "__name__"):
        return modules
    for module_info in pkgutil.walk_packages(
            path=package.__path__,
            prefix=package.__name__ + ".",
    ):
        module = importlib.import_module(module_info.name)
        modules.append(module)
    return modules


def _is_method(member) -> bool:
    return inspect.ismethod(member) or inspect.isfunction(member)


def _is_marker(member) -> bool:
    return isinstance(member, _Marker)


def _get_patched(
        fn: F,
        reference_injections: Dict[Any, Any],
        reference_closing: Dict[Any, Any],
) -> F:
    patched_object = PatchedCallable(
        original=fn,
        reference_injections=reference_injections,
        reference_closing=reference_closing,
    )

    if inspect.iscoroutinefunction(fn):
        patched = _get_async_patched(fn, patched_object)
    else:
        patched = _get_sync_patched(fn, patched_object)

    patched_object.patched = patched
    _patched_registry.register_callable(patched_object)

    return patched


def _is_fastapi_depends(param: Any) -> bool:
    return fastapi and isinstance(param, fastapi.params.Depends)


def _is_patched(fn) -> bool:
    return _patched_registry.has_callable(fn)


def _is_declarative_container(instance: Any) -> bool:
    return (isinstance(instance, type)
            and getattr(instance, "__IS_CONTAINER__", False) is True
            and getattr(instance, "declarative_parent", None) is None)


def _safe_is_subclass(instance: Any, cls: Type) -> bool:
    try:
        return issubclass(instance, cls)
    except TypeError:
        return False


class Modifier:

    def modify(
            self,
            provider: providers.ConfigurationOption,
            providers_map: ProvidersMap,
    ) -> providers.Provider:
        ...


class TypeModifier(Modifier):

    def __init__(self, type_: Type) -> None:
        self.type_ = type_

    def modify(
            self,
            provider: providers.ConfigurationOption,
            providers_map: ProvidersMap,
    ) -> providers.Provider:
        return provider.as_(self.type_)


def as_int() -> TypeModifier:
    """Return int type modifier."""
    return TypeModifier(int)


def as_float() -> TypeModifier:
    """Return float type modifier."""
    return TypeModifier(float)


def as_(type_: Type) -> TypeModifier:
    """Return custom type modifier."""
    return TypeModifier(type_)


class RequiredModifier(Modifier):

    def __init__(self) -> None:
        self.type_modifier = None

    def as_int(self) -> "RequiredModifier":
        self.type_modifier = TypeModifier(int)
        return self

    def as_float(self) -> "RequiredModifier":
        self.type_modifier = TypeModifier(float)
        return self

    def as_(self, type_: Type) -> "RequiredModifier":
        self.type_modifier = TypeModifier(type_)
        return self

    def modify(
            self,
            provider: providers.ConfigurationOption,
            providers_map: ProvidersMap,
    ) -> providers.Provider:
        provider = provider.required()
        if self.type_modifier:
            provider = provider.as_(self.type_modifier.type_)
        return provider


def required() -> RequiredModifier:
    """Return required modifier."""
    return RequiredModifier()


class InvariantModifier(Modifier):

    def __init__(self, id: str) -> None:
        self.id = id

    def modify(
            self,
            provider: providers.ConfigurationOption,
            providers_map: ProvidersMap,
    ) -> providers.Provider:
        invariant_segment = providers_map.resolve_provider(self.id)
        return provider[invariant_segment]


def invariant(id: str) -> InvariantModifier:
    """Return invariant modifier."""
    return InvariantModifier(id)


class ProvidedInstance(Modifier):

    TYPE_ATTRIBUTE = "attr"
    TYPE_ITEM = "item"
    TYPE_CALL = "call"

    def __init__(self) -> None:
        self.segments = []

    def __getattr__(self, item):
        self.segments.append((self.TYPE_ATTRIBUTE, item))
        return self

    def __getitem__(self, item):
        self.segments.append((self.TYPE_ITEM, item))
        return self

    def call(self):
        self.segments.append((self.TYPE_CALL, None))
        return self

    def modify(
            self,
            provider: providers.Provider,
            providers_map: ProvidersMap,
    ) -> providers.Provider:
        provider = provider.provided
        for type_, value in self.segments:
            if type_ == ProvidedInstance.TYPE_ATTRIBUTE:
                provider = getattr(provider, value)
            elif type_ == ProvidedInstance.TYPE_ITEM:
                provider = provider[value]
            elif type_ == ProvidedInstance.TYPE_CALL:
                provider = provider.call()
        return provider


def provided() -> ProvidedInstance:
    """Return provided instance modifier."""
    return ProvidedInstance()


class ClassGetItemMeta(GenericMeta):
    def __getitem__(cls, item):
        # Spike for Python 3.6
        if isinstance(item, tuple):
            return cls(*item)
        return cls(item)


class _Marker(Generic[T], metaclass=ClassGetItemMeta):

    __IS_MARKER__ = True

    def __init__(
            self,
            provider: Union[providers.Provider, Container, str],
            modifier: Optional[Modifier] = None,
    ) -> None:
        if _is_declarative_container(provider):
            provider = provider.__self__
        self.provider = provider
        self.modifier = modifier

    def __class_getitem__(cls, item) -> T:
        if isinstance(item, tuple):
            return cls(*item)
        return cls(item)

    def __call__(self) -> T:
        return self


class Provide(_Marker):
    ...


class Provider(_Marker):
    ...


class Closing(_Marker):
    ...


class AutoLoader:
    """Auto-wiring module loader.

    Automatically wire containers when modules are imported.
    """

    def __init__(self) -> None:
        self.containers = []
        self._path_hook = None

    def register_containers(self, *containers) -> None:
        self.containers.extend(containers)

        if not self.installed:
            self.install()

    def unregister_containers(self, *containers) -> None:
        for container in containers:
            self.containers.remove(container)

        if not self.containers:
            self.uninstall()

    def wire_module(self, module) -> None:
        for container in self.containers:
            container.wire(modules=[module])

    @property
    def installed(self) -> bool:
        return self._path_hook in sys.path_hooks

    def install(self) -> None:
        if self.installed:
            return

        loader = self

        class SourcelessFileLoader(importlib.machinery.SourcelessFileLoader):
            def exec_module(self, module):
                super().exec_module(module)
                loader.wire_module(module)

        class SourceFileLoader(importlib.machinery.SourceFileLoader):
            def exec_module(self, module):
                super().exec_module(module)
                loader.wire_module(module)

        class ExtensionFileLoader(importlib.machinery.ExtensionFileLoader):
            ...

        loader_details = [
            (SourcelessFileLoader, importlib.machinery.BYTECODE_SUFFIXES),
            (SourceFileLoader, importlib.machinery.SOURCE_SUFFIXES),
            (ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES),
        ]

        self._path_hook = importlib.machinery.FileFinder.path_hook(*loader_details)

        sys.path_hooks.insert(0, self._path_hook)
        sys.path_importer_cache.clear()
        importlib.invalidate_caches()

    def uninstall(self) -> None:
        if not self.installed:
            return

        sys.path_hooks.remove(self._path_hook)
        sys.path_importer_cache.clear()
        importlib.invalidate_caches()


def register_loader_containers(*containers: Container) -> None:
    """Register containers in auto-wiring module loader."""
    _loader.register_containers(*containers)


def unregister_loader_containers(*containers: Container) -> None:
    """Unregister containers from auto-wiring module loader."""
    _loader.unregister_containers(*containers)


def install_loader() -> None:
    """Install auto-wiring module loader hook."""
    _loader.install()


def uninstall_loader() -> None:
    """Uninstall auto-wiring module loader hook."""
    _loader.uninstall()


def is_loader_installed() -> bool:
    """Check if auto-wiring module loader hook is installed."""
    return _loader.installed


_patched_registry = PatchedRegistry()
_inspect_filter = InspectFilter()
_loader = AutoLoader()

# Optimizations
from ._cwiring import _get_sync_patched  # noqa
from ._cwiring import _async_inject  # noqa


# Wiring uses the following Python wrapper because there is
# no possibility to compile a first-type citizen coroutine in Cython.
def _get_async_patched(fn: F, patched: PatchedCallable) -> F:
    @functools.wraps(fn)
    async def _patched(*args, **kwargs):
        return await _async_inject(
            fn,
            args,
            kwargs,
            patched.injections,
            patched.closing,
        )
    return _patched



"""FactoryAggregate provider tests."""

from dependency_injector import providers, errors
from pytest import fixture, mark, raises

from .common import ExampleA, ExampleB


@fixture
def factory_a():
    return providers.Factory(ExampleA)


@fixture
def factory_b():
    return providers.Factory(ExampleB)


@fixture
def factory_type():
    return "default"


@fixture
def factory_aggregate(factory_type, factory_a, factory_b):
    if factory_type == "empty":
        return providers.FactoryAggregate()
    elif factory_type == "non-string-keys":
        return providers.FactoryAggregate({
            ExampleA: factory_a,
            ExampleB: factory_b,
        })
    elif factory_type == "default":
        return providers.FactoryAggregate(
            example_a=factory_a,
            example_b=factory_b,
        )
    else:
        raise ValueError("Unknown factory type \"{0}\"".format(factory_type))


def test_is_provider(factory_aggregate):
    assert providers.is_provider(factory_aggregate) is True


def test_is_delegated_provider(factory_aggregate):
    assert providers.is_delegated(factory_aggregate) is True


@mark.parametrize("factory_type", ["non-string-keys"])
def test_init_with_non_string_keys(factory_aggregate, factory_a, factory_b):
    object_a = factory_aggregate(ExampleA, 1, 2, init_arg3=3, init_arg4=4)
    object_b = factory_aggregate(ExampleB, 11, 22, init_arg3=33, init_arg4=44)

    assert isinstance(object_a, ExampleA)
    assert object_a.init_arg1 == 1
    assert object_a.init_arg2 == 2
    assert object_a.init_arg3 == 3
    assert object_a.init_arg4 == 4

    assert isinstance(object_b, ExampleB)
    assert object_b.init_arg1 == 11
    assert object_b.init_arg2 == 22
    assert object_b.init_arg3 == 33
    assert object_b.init_arg4 == 44

    assert factory_aggregate.factories == {
        ExampleA: factory_a,
        ExampleB: factory_b,
    }


def test_init_with_not_a_factory():
    with raises(errors.Error):
        providers.FactoryAggregate(
            example_a=providers.Factory(ExampleA),
            example_b=object(),
        )


@mark.parametrize("factory_type", ["empty"])
def test_init_optional_providers(factory_aggregate, factory_a, factory_b):
    factory_aggregate.set_providers(
        example_a=factory_a,
        example_b=factory_b,
    )
    assert factory_aggregate.providers == {
        "example_a": factory_a,
        "example_b": factory_b,
    }
    assert isinstance(factory_aggregate("example_a"), ExampleA)
    assert isinstance(factory_aggregate("example_b"), ExampleB)


@mark.parametrize("factory_type", ["non-string-keys"])
def test_set_factories_with_non_string_keys(factory_aggregate, factory_a, factory_b):
    factory_aggregate.set_providers({
        ExampleA: factory_a,
        ExampleB: factory_b,
    })

    object_a = factory_aggregate(ExampleA, 1, 2, init_arg3=3, init_arg4=4)
    object_b = factory_aggregate(ExampleB, 11, 22, init_arg3=33, init_arg4=44)

    assert isinstance(object_a, ExampleA)
    assert object_a.init_arg1 == 1
    assert object_a.init_arg2 == 2
    assert object_a.init_arg3 == 3
    assert object_a.init_arg4 == 4

    assert isinstance(object_b, ExampleB)
    assert object_b.init_arg1 == 11
    assert object_b.init_arg2 == 22
    assert object_b.init_arg3 == 33
    assert object_b.init_arg4 == 44

    assert factory_aggregate.providers == {
        ExampleA: factory_a,
        ExampleB: factory_b,
    }


def test_set_providers_returns_self(factory_aggregate, factory_a):
    assert factory_aggregate.set_providers(example_a=factory_a) is factory_aggregate


@mark.parametrize("factory_type", ["empty"])
def test_init_optional_factories(factory_aggregate, factory_a, factory_b):
    factory_aggregate.set_factories(
        example_a=factory_a,
        example_b=factory_b,
    )
    assert factory_aggregate.factories == {
        "example_a": factory_a,
        "example_b": factory_b,
    }
    assert isinstance(factory_aggregate("example_a"), ExampleA)
    assert isinstance(factory_aggregate("example_b"), ExampleB)


@mark.parametrize("factory_type", ["non-string-keys"])
def test_set_factories_with_non_string_keys(factory_aggregate, factory_a, factory_b):
    factory_aggregate.set_factories({
        ExampleA: factory_a,
        ExampleB: factory_b,
    })

    object_a = factory_aggregate(ExampleA, 1, 2, init_arg3=3, init_arg4=4)
    object_b = factory_aggregate(ExampleB, 11, 22, init_arg3=33, init_arg4=44)

    assert isinstance(object_a, ExampleA)
    assert object_a.init_arg1 == 1
    assert object_a.init_arg2 == 2
    assert object_a.init_arg3 == 3
    assert object_a.init_arg4 == 4

    assert isinstance(object_b, ExampleB)
    assert object_b.init_arg1 == 11
    assert object_b.init_arg2 == 22
    assert object_b.init_arg3 == 33
    assert object_b.init_arg4 == 44

    assert factory_aggregate.factories == {
        ExampleA: factory_a,
        ExampleB: factory_b,
    }


def test_set_factories_returns_self(factory_aggregate, factory_a):
    assert factory_aggregate.set_factories(example_a=factory_a) is factory_aggregate


def test_call(factory_aggregate):
    object_a = factory_aggregate("example_a", 1, 2, init_arg3=3, init_arg4=4)
    object_b = factory_aggregate("example_b", 11, 22, init_arg3=33, init_arg4=44)

    assert isinstance(object_a, ExampleA)
    assert object_a.init_arg1 == 1
    assert object_a.init_arg2 == 2
    assert object_a.init_arg3 == 3
    assert object_a.init_arg4 == 4

    assert isinstance(object_b, ExampleB)
    assert object_b.init_arg1 == 11
    assert object_b.init_arg2 == 22
    assert object_b.init_arg3 == 33
    assert object_b.init_arg4 == 44


def test_call_factory_name_as_kwarg(factory_aggregate):
    object_a = factory_aggregate(
        factory_name="example_a",
        init_arg1=1,
        init_arg2=2,
        init_arg3=3,
        init_arg4=4,
    )
    assert isinstance(object_a, ExampleA)
    assert object_a.init_arg1 == 1
    assert object_a.init_arg2 == 2
    assert object_a.init_arg3 == 3
    assert object_a.init_arg4 == 4


def test_call_no_factory_name(factory_aggregate):
    with raises(TypeError):
        factory_aggregate()


def test_call_no_such_provider(factory_aggregate):
    with raises(errors.NoSuchProviderError):
        factory_aggregate("unknown")


def test_overridden(factory_aggregate):
    with raises(errors.Error):
        factory_aggregate.override(providers.Object(object()))


def test_getattr(factory_aggregate, factory_a, factory_b):
    assert factory_aggregate.example_a is factory_a
    assert factory_aggregate.example_b is factory_b


def test_getattr_no_such_provider(factory_aggregate):
    with raises(errors.NoSuchProviderError):
        factory_aggregate.unknown


def test_factories(factory_aggregate, factory_a, factory_b):
    assert factory_aggregate.factories == dict(
        example_a=factory_a,
        example_b=factory_b,
    )


def test_deepcopy(factory_aggregate):
    provider_copy = providers.deepcopy(factory_aggregate)

    assert factory_aggregate is not provider_copy
    assert isinstance(provider_copy, type(factory_aggregate))

    assert factory_aggregate.example_a is not provider_copy.example_a
    assert isinstance(factory_aggregate.example_a, type(provider_copy.example_a))
    assert factory_aggregate.example_a.cls is provider_copy.example_a.cls

    assert factory_aggregate.example_b is not provider_copy.example_b
    assert isinstance(factory_aggregate.example_b, type(provider_copy.example_b))
    assert factory_aggregate.example_b.cls is provider_copy.example_b.cls


@mark.parametrize("factory_type", ["non-string-keys"])
def test_deepcopy_with_non_string_keys(factory_aggregate):
    provider_copy = providers.deepcopy(factory_aggregate)

    assert factory_aggregate is not provider_copy
    assert isinstance(provider_copy, type(factory_aggregate))

    assert factory_aggregate.factories[ExampleA] is not provider_copy.factories[ExampleA]
    assert isinstance(factory_aggregate.factories[ExampleA], type(provider_copy.factories[ExampleA]))
    assert factory_aggregate.factories[ExampleA].cls is provider_copy.factories[ExampleA].cls

    assert factory_aggregate.factories[ExampleB] is not provider_copy.factories[ExampleB]
    assert isinstance(factory_aggregate.factories[ExampleB], type(provider_copy.factories[ExampleB]))
    assert factory_aggregate.factories[ExampleB].cls is provider_copy.factories[ExampleB].cls


def test_repr(factory_aggregate):
    assert repr(factory_aggregate) == (
        "<dependency_injector.providers."
        "FactoryAggregate({0}) at {1}>".format(
            repr(factory_aggregate.factories),
            hex(id(factory_aggregate)),
        )
    )



from httpx import ASGITransport, AsyncClient
from pytest import fixture, mark
from pytest_asyncio import fixture as aio_fixture

# Runtime import to avoid syntax errors in samples on Python < 3.5 and reach top-dir
import os
_SAMPLES_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        "../samples/",
    )),
)
import sys
sys.path.append(_SAMPLES_DIR)


from wiringfastapi import web


@aio_fixture
async def async_client():
    client = AsyncClient(transport=ASGITransport(app=web.app), base_url="http://test")
    yield client
    await client.aclose()


@mark.asyncio
async def test_depends_marker_injection(async_client: AsyncClient):
    class ServiceMock:
        async def process(self):
            return "Foo"

    with web.container.service.override(ServiceMock()):
        response = await async_client.get("/")

    assert response.status_code == 200
    assert response.json() == {"result": "Foo"}


@mark.asyncio
async def test_depends_injection(async_client: AsyncClient):
    response = await async_client.get("/auth", auth=("john_smith", "secret"))
    assert response.status_code == 200
    assert response.json() == {"username": "john_smith", "password": "secret"}



"""Self provider tests."""

from dependency_injector import providers, containers


def test_is_provider():
    assert providers.is_provider(providers.Self()) is True


def test_call_object_provider():
    container = containers.DeclarativeContainer()
    assert providers.Self(container)() is container


def test_set_container():
    container = containers.DeclarativeContainer()
    provider = providers.Self()
    provider.set_container(container)
    assert provider() is container


def test_set_alt_names():
    provider = providers.Self()
    provider.set_alt_names({"foo", "bar", "baz"})
    assert set(provider.alt_names) == {"foo", "bar", "baz"}


def test_deepcopy():
    provider = providers.Self()

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider, providers.Self)


def test_deepcopy_from_memo():
    provider = providers.Self()
    provider_copy_memo = providers.Provider()

    provider_copy = providers.deepcopy(provider, memo={id(provider): provider_copy_memo})

    assert provider_copy is provider_copy_memo


def test_deepcopy_overridden():
    provider = providers.Self()
    overriding_provider = providers.Provider()

    provider.override(overriding_provider)

    provider_copy = providers.deepcopy(provider)
    overriding_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert isinstance(provider, providers.Self)

    assert overriding_provider is not overriding_provider_copy
    assert isinstance(overriding_provider_copy, providers.Provider)


def test_repr():
    container = containers.DeclarativeContainer()
    provider = providers.Self(container)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "Self({0}) at {1}>".format(repr(container), hex(id(provider)))
    )
